/*** tty_raw.c
***/
/****  04.5.21  P.345 in UNIXץߥ   ***/

#include "lisp.h"
#include <termios.h>

#include <unistd.h>

#include <stdio.h>
#include <signal.h>
#include <sys/ioctl.h>

#define LINEMAX 4096
#define HISBUFFMAX 4096
#define BACKSPACE 8
#define DELETE 127      /**** redhat     Backspce key is delete!!!***/

extern int input_line,input_flag;
extern int top_visible;
extern void timer_handler();
extern int stdin_ready();

/***
   tty_raw(0)    non-canonical mode
      return  -1  for block device
   tty_reset(0)   reset to original
***/
   
static struct termios save_termios;
static int ttysavefd=-1;
static enum{RESET,RAW}ttystate=RESET;

unsigned char linebuff[LINEMAX+3];
static unsigned char history_buff[HISBUFFMAX+4];

int linepos=0,historypos=0;

int win_xsize=0,win_ysize=0,win_change;

static unsigned char cbuf[2];
static unsigned int nonwhite;

/*** peek  ***/
int tty_peek(){
  Reg ptr x,lpath;
  Reg int c,n,s;
  c=line_raw_inkey();
  if(c==SEOF) return SEOF;
  if((c<128)||!kanji_mode){
     ungetc_buff(c);
     return c;
   }
  n=raw_inkey();
  if(n==SEOF){
     ungetc_buff(c);
     return c;
   }
  if(kanji_mode==3){
    s=raw_inkey();
    ungetc_buff(s);
    ungetc_buff(n);
    ungetc_buff(c);
    return (c<<16)+(n<<8)+s;  /*** utf8  **/
   }
  ungetc_buff(n);
  ungetc_buff(c);
  return (c<<8)+n;
}

/***  tty  is white?  **/
int buffer_white(){
  int n;
  lfflush(STDOUT);
  n=buffer_white_main();
  return n;
}

int buffer_white_main(){
  ptr x,lpath;
  Reg int n,c;
   /** printf("buffer_white enter\n"); **/
      c=line_raw_inkey();
      if(c==SEOF) return YES;
  if(c&&c!='\n'&&c!=' '){
        nonwhite=c; 
      /***  because  if we store c , then readchar return
                       c   before enter tty-read . c is lost for tty-read! ***/
        /** ungetc_buff(c);  ***/
       return NO;
   }
    return YES; /*** drop top of space \n  ***/
}


int tty_raw(int fd)   /***  into raw stdin mode ***/
{
  struct termios buf;
  int vdisable;
  int get_win_size();
   if(ttystate==RAW) return 1;
        /*** P.321 **/
  #ifndef WIN32
   /**** error occur ????  ****/
   if((vdisable=fpathconf(fd,_PC_VDISABLE))<0){
        /** printf("tty 0\n"); **/
        lwritestr(STDERR,"tty 0\n");
        return 0;
   }
 #endif
   if(tcgetattr(fd,&save_termios)<0){
        /*** printf(" Not tty  \n"); ***/
        lwritestr(STDMSG," Not tty  \n");
      return 0;
    }
   buf=save_termios;
          /***  if you want  noecho then use ~(ECHO|ICANON)   ***/
   /*** buf.c_lflag&=~ICANON;   ***  non canonical  **/
   buf.c_lflag&=~(ECHO|ICANON);   /*** no echo && non canonical  **/
   buf.c_cc[VMIN]=0; /***  0 byte or 1 ***/
   buf.c_cc[VTIME]=0; /*** no wait !   if set 1   0.1 sec wait ***/
  
   buf.c_cc[VSTOP]=vdisable;   /***  STOP(CTRL-S) disable ***/
   buf.c_cc[VSTART]=vdisable;   /***  START(CTRL-Q) disable ***/
  if(tcsetattr(fd,TCSAFLUSH,&buf)<0)
         return 0;
   ttystate=RAW;
   ttysavefd=fd;
   /*** get window size ***/
   signal(SIGWINCH,get_win_size);
   /** get_win_size(); **/
   win_change=0;
   return 1; /*** OK  ***/
}

get_win_size(){
 struct winsize size;
  win_change=-1;
  if(ioctl(0,TIOCGWINSZ,(char *)&size)<0) return -1;
  /** printf("%d rows, %d columnds\n",size.ws_row,size.ws_col); ***/
  win_xsize=size.ws_col;
  win_ysize=size.ws_row;
  win_change=1;
  ATOMVAL(a_win_xsize)=MAKESINT(win_xsize);
  ATOMVAL(a_win_ysize)=MAKESINT(win_ysize);
  ATOMVAL(a_win_change)=a_t;
}

  
int tty_reset(int fd)
{
   if(ttystate!=RAW) return 0;
  if(tcsetattr(fd,TCSAFLUSH,&save_termios)<0)
         return -1;
   ttystate=RESET;
   return 0; /*** OK  ***/
}
/***  tty_ready  ***/
/***  return NO  Not or YES  ***/
int tty_ready()
{
 Reg ptr lpath,x;
  lpath=a_keyinput;
     x=CAR(CDDR(lpath));
     if(x!=SEOF){
        return YES;
     }
  if(linepos>0){
        return YES;
  }
   if(stdin_ready()) return YES;
  return NO;
}


/*** ungetc,linepos,else raw_inkey  return SEOF if not ready  **/
int line_raw_inkey(){
 Reg ptr lpath,x;
 Reg int ct,c;
    lpath=a_keyinput;
     x=CAR(CDDR(lpath));
     if(x!=SEOF){
        CAR(CDDR(lpath))=SEOF;
        return CHAR_CODE(x);
     }
  if(linepos>0){
      c= read_char_buff();
      /**  printf("raw inkey buffer linepos %d %c ",linepos,c); **/
       return c;
  }
    return raw_inkey(); 
}

int raw_inkey(){
 Reg int ct,c;
 if(stdin_ready()){
     ct=read(0,cbuf,1);
     if(ct>0){
        c=cbuf[0];
       /** printf("raw inkey read linepos %d %c ",linepos,c);***/
       return c;
     }
   }
  return SEOF; /***  not ready  ***/
}

int raw_inkey_wait(){
  int n;
  n=SEOF;
  while(n==SEOF){
   n=raw_inkey();
   ss_update();
        /*** if(n!=SEOF) printf("raw-inkey-wait %x  %c",n,n); ***/
#ifdef USE_TK
   if(top_visible) {
       /** printf("raw-wait\n"); **/
       update();
      }
#endif
  }
  return n;
}

memcpydown_s(char *dest,char *s,int size)
{
  Reg int c;
    while(size--) *dest++=*s++;
    *dest=0;
}

memcpydown(char *dest,char *s)
{
  Reg int c;
    while(c=*s++) *dest++=c;
    *dest=0;
}

memcpyup_s(char *dest,char *s,int size)
{
  Reg int c,n;
    s+=size;
    dest+=size;
    while(size--) *--dest=*--s;
}

memcpyup(char *dest,char *s)
{
  Reg int c,n;
    n=0;
    while(c=*s++) n++;
    n++;
    dest+=n;
    while(n--) *--dest=*--s;
}    

int read_char_buff(){
  int c;
  if(linepos==0) return SEOF;
  c=linebuff[0];
   memcpydown_s(linebuff,linebuff+1,linepos--);
  return c;
}

int inkey(int echo,int wait){
  ptr x,lpath;
  int n=0,c;
  c=line_raw_inkey();
  if(c!=SEOF){
     if(echo) goto echo;
     return c;
  }
  if(wait){
         /** printf("inkey wait top_visible %d\n",top_visible); ***/
          c=raw_inkey_wait();
      /** input_buffer(); ***/
  }
  else{
        c=raw_inkey();
  }
  if(c==SEOF) return c;
echo:
  if(echo){
        lwritec(STDOUT,c);
        lfflush(STDOUT);
  }
  return c;  
}

input_buffer(){
  int n=1;
    while((n!=SEOF)&&( linepos<LINEMAX-3)){
          n=raw_inkey();
          if(n!=SEOF)
            store_char(n);
    }
}

int tty_getchar()
{
  Reg int n,m;
  m=n=tty_getchar_sub();
  if(input_flag){
    if(historypos<HISBUFFMAX){
      if(n>255){
         if(n>0xffff){
         history_buff[historypos++]=(n>>16); /** utf8 **/
         }
         history_buff[historypos++]=(n>>8);
         n&=0xff;
        }
      history_buff[historypos++]=n;
      history_buff[historypos]=0;
    }
  }
 return m;
}

int tty_getchar_sub()
{
  ptr x,lpath;
  int n=0;
   lpath=a_keyinput;
  if(input_flag&&(x=ATOMVAL(a_tty_read))!=a_undef){
       return tty_readchar(lpath,x);
  }
   x=CAR(CDDR(lpath));
   if(x!=SEOF){
      /*** printf("tty_getchar %c\n",x); ***/
      CAR(CDDR(lpath))=SEOF;
      return CHAR_CODE(x);
   }
  if(linepos>0){
      /*** printf("tty_getchar linepos %d\n",linepos); ***/
       return  read_char_buff();
  }
  simple_line_get();
  return  read_char_buff();
}

int tty_readchar(ptr lpath,ptr func){
  ptr x;
  int n=0;
    x=CAR(CDDR(lpath));
    if(x!=SEOF){
      CAR(CDDR(lpath))=SEOF;
      ungetc_buff(CHAR_CODE(x));
   }
   if(nonwhite){
      ungetc_buff(nonwhite);
      /*** printf("nonwhite ungetc %c\n",nonwhite);***/
      nonwhite=0;
      }
    x=funcall(func,0,a_nil);
    n=CHAR_CODE(x);
    /*** printf("tty_readchar %x %c\n",n,n); ***/
/***
    if(n>255){
         ungetc_buff(n&0xff);
         return (n>>8);
     }
***/
     return n;
}


simple_line_get()
{
  Reg int c;
      /**  printf("simple nonwhite %c\n",nonwhite); ***/
     if(c=nonwhite){
        nonwhite=0;
        echo_char(c);
        if(c=='\n') return;
      }
   if((linepos>0)&&(linebuff[linepos-1]=='\n')) return; 
   while(1){
          c=raw_inkey_wait();
       /*** printf("simple raw_inkey %c\n",c); ***/
     echo_char(c);
    if(c=='\n') return;
    if(linepos==LINEMAX) return;
  }
}

store_char(int c){
   linebuff[linepos++]=c;
   linebuff[linepos]=0;
}

ungetc_buff(int c){
   if(c==-1) return;
   if(c>255){
      if(kanji_mode==3){
       memcpyup_s(linebuff+3,linebuff,linepos);
       linepos+=3;
       linebuff[2]=c&0xff;
       linebuff[1]=(c>>8);
       linebuff[0]=(c>>16);
       return;
       }
      else{
       memcpyup_s(linebuff+2,linebuff,linepos);
       linepos+=2;
       linebuff[1]=c&0xff;
       linebuff[0]=(c>>8);
       return;
       }
     }
   memcpyup_s(linebuff+1,linebuff,linepos++);
   linebuff[0]=c;
}   

echo_char(int c){
  Reg ptr out=STDOUT;
   if((c==BACKSPACE)||(c==DELETE)){
         if(linepos>0){
               c=BACKSPACE;
             lwritec(out,c);
             lwritec(out,' ');
             lwritec(out,c);
             lfflush(out);
             linebuff[--linepos]=0;
          }
     }
    else{
       lwritec(out,c);
       lfflush(out);
       linebuff[linepos++]=c;
       linebuff[linepos]=0;
    }
}

static char *space_off(char *p,int len)
{
  char *q;
  int c;
  q=p+len;
  while(1){
   c=*--q;
   if((c=='\n')||(c==' '))
        *q=0;
   else
      break;
  }
  while(1){
   c=*p;
   if((c=='\n')||(c==' '))
      p++;
   else
      break;
  }
 return p;
}    

line_history(int line){
  ptr v,*p,x;
  int size;
  char *u;
  u=space_off(history_buff,historypos);
  /*** printf("%d %s \n",line,history_buff); ***/
  v=ATOMVAL(a_linehis);
  p=get_vector(v,&size);
  p+=size;
  while(--size){
      --p;
      *p=*(p-1);
  }
  x=CMAKESTR(u);
  p=get_vector(v,&size);
  *p=x;
  historypos=0;
  history_buff[0]=0;
} 

