#include "lisp.h"
#include <math.h>

// sqrt 2 x>= 1.5
//  n=25  err 1.0e-10
static double F(double x){
  double s,a;
  int n=25;
   a=x*x+4*26;  //x^2+4(n+1)
   s=(sqrt(a)-x)/2;
   while(n>0){
     s=n/(x+s);
     n--;
  }
  return  s;
}


static double erfc_F(double x){
  double s;
  s=sqrt(2)*x;
  s=1/(s+F(s));
   if(x*x>320.0) return 0.0; 
  return sqrt(2/M_PI)*s*exp(-x*x);
}
// x>=2.13=1.5*sqrt(2)
double Nerf(double x);
double Nerfc(double x){
  double s;
  if(x<0.0) return 0.5+Nerf(-x);
  if(x<2.13) return 0.5*erfc(x/sqrt(2));
   if(x*x>640.0) return 0.0; 
  s=(x+F(x))*sqrt(2*M_PI);
  return exp(-x*x*0.5)/s;
}
double Nerf(double x){
  double s;
  return 0.5*erf(x/sqrt(2));
}

double erf(double x){
  double s=1.0,x2,a,g=1.0;
  int n=1;
  if(x<0.0) return -erf(-x);
  if(x>=1.5) return 1-erfc_F(x);
  x2=x*x;
 while(1){
  g=-x2*g/n;
  a=g/(2*n+1);
  s+=a;
  if(fabs(a/s)<=1.0E-15) break;
  n++;
  }
 return 2*x*s/sqrt(M_PI);
}
double erfc(double x){
   if(x<0.0) return 1+erf(-x);
  if(x<1.5) return 1-erf(x);
  return erfc_F(x);
}

/***  Only Real ***/
ptr ERF(y)
  Reg ptr y;
{
Reg double x;
   if(REALP(y))
      {
          x=get_float(y);
        return MAKEDBL(erf(x));
       }
   error(E_notnum,y);
}
ptr ERFC(y)
  Reg ptr y;
{
Reg double x;
   if(REALP(y))
      {
          x=get_float(y);
        return MAKEDBL(erfc(x));
       }
   error(E_notnum,y);
}
ptr NERF(y)
  Reg ptr y;
{
Reg double x;
   if(REALP(y))
      {
          x=get_float(y);
        return MAKEDBL(Nerf(x));
       }
    error(E_notnum,y);
}
ptr NERFC(y)
  Reg ptr y;
{
Reg double x;
   if(REALP(y))
      {
          x=get_float(y);
        return MAKEDBL(Nerfc(x));
       }
    error(E_notnum,y);
}

ptr s_erf(int narg,ptr env)

 {
  checknarg(narg-1);
    return ERF(*(stack-1));
 }

ptr s_erfc(int narg,ptr env)

 {
  checknarg(narg-1);
    return ERFC(*(stack-1));
 }

ptr s_Nerf(int narg,ptr env)

 {
  checknarg(narg-1);
    return NERF(*(stack-1));
 }

ptr s_Nerfc(int narg,ptr env)

 {
  checknarg(narg-1);
    return NERFC(*(stack-1));
 }



/****
int main(){
  double x,s,a;
  for(x=-2.0;x<4;x+=0.1){
   printf("x=%f  erfc=%f\n",x,erfc(x));
  }
  for(x=-2.0;x<4;x+=0.1){
   printf("x=%f  Nerfc=%f\n",x,Nerfc(x));
  }
  for(x=-1.0;x<3.05;x+=0.1){
   printf("x=%f  erf=%f\n",x,erf(x));
  }
  return 0;
}
****/ 
