/****   Math   functions 
   f_math.c 
***/
#define  F_MATH_C

#include "lisp.h"
#include <stdio.h>
#include <math.h>

extern mpq_t mpqwork1, mpqwork2;
extern mpz_t mpzwork1, mpzwork2;
extern ptr genmul(),gendiv(),genadd(),gensub();
extern ptr fixdiv();
extern ptr DBL0;

#define  getfloat(x)  (FLONUM(x))

extern double get_float();

/***  cosh,sinh conflict MACOSX system   -->  ss_cosh,  ss_sinh***/

ptr s_float(int narg,ptr env)
    
 {
    /*   expr 1  ,  -->double  */
        return(_float(*(stack-1)));
 }

int gsign(x)
    Reg ptr x;
 {
   /****   sign of x  :  return 1 if x is negative ****/
    if(SINTP(x)) 
           return ((long)x<0);
    else if(BIGINTP(x))
     {
       return (mpz_sgn(BIGNUM(x))==-1);
     }
    else if(RATIOP(x))
     {
       return (mpq_sgn(RATIONAL(x))==-1);
     }
    else if(DOUBLEP(x))
        return (FLONUM(x)<0.0);
        /** return ( (int)CAR(x)<0);    sign bit is MSB of double!!! ***/
   error(E_notreal,x);
 }

int zerop(x)
    Reg ptr x;
 {
   if(SINTP(x))
       return x==ZERO;
   else if(DOUBLEP(x)){
      /***   =との整合性  ***/
     return  gen_eq(x,ZERO);
   }
      /*** return (FLONUM(x)==0.0);  ***/
   else if(COMPLEXP(x))
      return (zerop(CAR(x))&&zerop(CDR(x)));
   else if((int)ATTR(x)&NUMBER)
      return NO;
   else
     error(E_notnum,x);
 }

ptr minus(x)
    Reg ptr x;
 {
   /****   (- x) ****/
   Reg ptr n;
   long m;
    if(SINTP(x)){ 
       /***  2014.5.10  Bug fix  **/
#ifdef SYSTEM64
        m=(long)x;
      if((m&0xffffffff)==0x80000001)
           return bignum((long)0x20000000);
#else
        m=(long)x;  //32 bit
      if(m==0x80000001) 
           return bignum((long)0x20000000);
#endif
       return (ptr)((~m)+3);
     }
    else if(BIGINTP(x))
     {
       return longneg(x);
     }
    else if(RATIOP(x))
     {
       psh(x);
       mpq_neg(mpqwork1,RATIONAL(x));
       x= mpq2rational(mpqwork1);
       pull();
       return x;
     }
    else if(DOUBLEP(x))
      {
        //2018.3.10
       psh(x);
       x=MAKEDBL(-FLONUM(x));
       pull();
       return x;
      }
    else if(COMPLEXP(x))
      {
        psh(x);
        n=minus(CAR(x));
        psh(n);
        x=minus(CDR(x));
        x=SCONS(n,x);  //2018.3.10
        ATTR(x)=COMPLEX;
        pull();
        pull();
       return x;
      }
   error(E_notnum,x);
 }

int check_big(p)
  Reg ptr p;
{
  /****   get 32 bit integer from  BIGINT  ****/
  /****   if exceed 32bit return 0  ***/
  /***   see check16   **/
  Reg int n;
    n=0;
    psh(p);
   /** if(mpz_fits_slong_p(BIGNUM(p))){    This is 64 bit in 64 system! Not int!**/
   if(mpz_fits_sint_p(BIGNUM(p))){
      n= mpz_get_si(BIGNUM(p));
    }
    pull();
  return n;
}

int gen_eq(x,y)
    Reg ptr x,y;
 {
    /****  =,  /=    ***/
   Reg int s;
   Reg double v,z,u;
   
    if(x==y) return YES;
    if(SINTP(x))
        {
           if(SINTP(y)) return NO;
           if(DOUBLEP(y))
             {
               v=(double)SINT_NUMB(x);
               /*** return (FLONUM(y)==v); ***/
               return gen_eq(MAKEDBL(v),y);
             }
           if(BIGINTP(y)||RATIOP(y)) return NO;
           if(COMPLEXP(y))
            {
              if(gen_eq(x,CAR(y))) return gen_eq(CDR(y),ZERO);
              return NO;
             }
           error (E_notnum,y);
       }
   if(SINTP(y)) return gen_eq(y,x); /*** needs , because COMPLEXP(y) below **/
   if(COMPLEXP(x))
      {
        if(COMPLEXP(y))
         {
            if(gen_eq(CAR(x),CAR(y))) return gen_eq(CDR(x),CDR(y));
            else
              return NO;
          }
        if(gen_eq(ZERO,CDR(x))) return gen_eq(CAR(x),y);
         return NO;
      }
  if(COMPLEXP(y))
    {
      if(gen_eq(ZERO,CDR(y))) 
         y=CAR(y);
       else
          return NO;
     }
   /***  x,y are  not COMPLEX ***/
   if(DOUBLEP(x))
      {
         /** printf("gen_eq dbl x %f y %f x-y %e\n",FLONUM(x),get_float(y),FLONUM(x)-get_float(y)); **/
        /*** return (FLONUM(x)==get_float(y)); ***/
        v=FLONUM(x);
        z=get_float(y);
        u=fabs(v-z);
        if(u<2.9e-39) return YES;   /**** this is nearly 0.0 in float sense ****/
       
        /***  if(v==0.0)
             return (fabs(z)<2.9e-39); 
        ***/
        v=fabs(v)+fabs(z);
        return ((u/v)<2.3e-16);    /*** double の仮数部は、52 bit ***/
        
       /***   return (0.0==(FLONUM(x)-get_float(y)));    this is no good ***/
       }
   if(BIGINTP(x)) 
     {
       if(SINTP(y)||RATIOP(y)) 
       {
        return NO;
        }
       if(DOUBLEP(y))
         {
        return (FLONUM(y)==get_float(x));
         }
       if(BIGINTP(y)) 
         {
           psh(x);
           psh(y);
           s= (mpz_cmp(BIGNUM(x),BIGNUM(y))==0);
           stack-=2;
           return s;
         }
         error (E_notnum,y);
    }
    if(RATIOP(x))
        {
           if(SINTP(y)||BIGINTP(y)) return NO;
           if(DOUBLEP(y))
             {
                  return (FLONUM(y)==get_float(x));
             }
           if(RATIOP(y)) 
             {
               psh(x);
               psh(y);
               s= (mpq_cmp(RATIONAL(x),RATIONAL(y))==0);
               stack-=2;
               return s;
             }
           error (E_notnum,y);
       }
   error (E_notnum,y);
 }

int gen_order(x,y)
    Reg ptr x,y;
    /*   x>y 1, x==y 0,  x<y  -1  */
 {
   Reg int n,s;
   Reg ptr p;
   
    if(x==y) return 0;
    if(SINTP(x))
       {
         n=getnum(x);
         if(SINTP(y)) 
              return int_order(n,getnum(y));
         if(DOUBLEP(y))
              return double_order((double)n,FLONUM(y));
         if(BIGINTP(y)) 
          {
            if(s=check_big(y)) return int_order(n,s);
            return (gsign(y))?1:-1;
          }
         if(RATIOP(y))
          {
            psh(y);
            n=mpq_cmp_si(RATIONAL(y),SINT_NUMB(x),1);
            pull();
            return -n;
          }
        error (E_notnum,y);
       }
    if(DOUBLEP(x))
      {
       return double_order(FLONUM(x),get_float(y));
       }
    if(BIGINTP(x)) 
      {
       if(SINTP(y)) 
       {
        if(n=check_big(x)) return int_order(n,SINT_NUMB(y));
         return (gsign(x))?-1:1;
       }
       if(DOUBLEP(y))
       {
         return double_order(get_float(x),FLONUM(y));
       }
      if(BIGINTP(y)) 
       {
           psh(x);
           psh(y);
           s= mpz_cmp(BIGNUM(x),BIGNUM(y));
           stack-=2;
           return s;
       }
         if(RATIOP(y))
          {
            psh(x);
            psh(y);
            mpq_set_z(mpqwork1,BIGNUM(x));
            s= mpq_cmp(mpqwork1,RATIONAL(y));
            stack-=2;
            return s;
          }
       error (E_notnum,y);
    }
    if(RATIOP(x))
        {
           if(SINTP(y)){
               psh(x);
               s= mpq_cmp_si(RATIONAL(x),SINT_NUMB(y),1);
               pull();
               return s;
               }
           else if(BIGINTP(y)) {
               psh(x);
               psh(y);
               mpq_set_z(mpqwork1,BIGNUM(y));
               s= mpq_cmp(RATIONAL(x),mpqwork1);
               stack-=2;
               return s;
            }
           else if(DOUBLEP(y))
             {
               return double_order(get_float(x),FLONUM(y));
             }
           else if(RATIOP(y)) 
             {
               psh(x);
               psh(y);
               s= mpq_cmp(RATIONAL(x),RATIONAL(y));
               stack-=2;
               return s;
             }
       }
    error (E_notreal,x);
 }
int int_order(n,m)
  Reg int n,m;
{
  if(n==m) return 0;
  if(n>m)  return 1;
  else
    return -1;
}
int double_order(n,m)
  Reg double n,m;
{
  if(n==m) return 0;
  if(n>m)  return 1;
  else
    return -1;
}
/*****
ptr g_eq(int narg,ptr env)
    
 {
  checknarg(narg-2);
  return gen_eq(*(stack-2),*(stack-1))?a_t:a_f; 
 }
ptr g_gt(int narg,ptr env)
    
 {
  checknarg(narg-2);
         return (gen_order(*(stack-2),*(stack-1))>0)?a_t:a_f;
 }
ptr g_lt(int narg,ptr env)
    
 {
  checknarg(narg-2);
         return (gen_order(*(stack-2),*(stack-1))<0)?a_t:a_f;
 }
****/
ptr g_ne(int narg,ptr env)
    
 {
  checknarg(narg-2);
         return (gen_eq(*(stack-2),*(stack-1))==0)?a_t:a_f;
 }
/****
ptr g_ge(int narg,ptr env)
    
 {
  checknarg(narg-2);
         return (gen_order(*(stack-2),*(stack-1))>=0)?a_t:a_f;
 }
ptr g_le(int narg,ptr env)
    
 {
  checknarg(narg-2);
         return (gen_order(*(stack-2),*(stack-1))<=0)?a_t:a_f;
 }
****/

ptr g_max(int narg,ptr env)
    /*  lexpr version */
 {
    Reg ptr x,pmax,*free;
    if(narg==0) error(E_shortarg,a_nil);
    free=stack-narg;
    pmax=*free++;
    narg--;
   while(narg)
   {
    x=*free++;
    if(gen_order(pmax,x)<0) pmax=x;
    narg--;
   }
    return pmax;
 }

ptr g_min(int narg,ptr env)
    /*  lexpr version */
 {
    Reg ptr *free,x,pmin;
    if(narg==0) error(E_shortarg,a_nil);
      free=stack-narg;
    pmin=*free++;
    narg--;
   while(narg)
   {
    x=*free++;
    if(gen_order(pmin,x)>0) pmin=x;
    narg--;
   }
    return pmin;
 }

ptr g_add1(int narg,ptr env)
 {
   /*  any int,double */
    /* register ptr x=*(stack-1),p; */
  checknarg(narg-1);
      return genadd(*(stack-1),ONE);
    
 }

ptr g_sub1(int narg,ptr env)
 {
  checknarg(narg-1);
      return gensub(*(stack-1),ONE);
 }

ptr ABS(x)
    ptr x;
 {
      if(REALP(x))
        {
          if(gsign(x)) return minus(x);
          return x;
        }
      if(COMPLEXP(x))
        {
          return MAKEDBL(sqrt(get_float(norm2(x))));
        }
   error(E_notnum,x);
 }

ptr g_abs(int narg,ptr env)
 {
  checknarg(narg-1);
   return ABS(*(stack-1));
 }

ptr g_zerop(int narg,ptr env)
    
 {
    Reg ptr x=*(stack-1);
  checknarg(narg-1);
    return PRED(zerop(*(stack-1)));
   
 }

ptr g_minusp(int narg,ptr env)
    
 {
  checknarg(narg-1);
    return PRED(gsign(*(stack-1)));
 }

ptr g_plusp(int narg,ptr env)
    
 {
    Reg ptr x=*(stack-1);
  checknarg(narg-1);
    if(zerop(x)||gsign(x)) return a_f;
   return a_t;
 }

ptr g_sign(int narg,ptr env)
    
 {
   Reg ptr x=*(stack-1);
  checknarg(narg-1);
   if(zerop(x)) return ZERO;
   if(gsign(x)) return MINUSONE;
   return ONE;
 }

/*  sin,cos,tan,acos,asin,atan,exp,sqrt */
ptr ATAN(),LOG(),EXP(),CIS();
ptr SINH(),COSH(), TANH(), ASINH(), ACOSH(), LOGRT();
static ptr mul_imag(y)
  Reg ptr y;
 {
    /*** return iy ***/
  if(REALP(y))
    return complex(ZERO,y);
  if(COMPLEXP(y))
    return complex(minus(CDR(y)),CAR(y));
  error(E_notnum,y);
 }

static ptr mul_imag2(y)
  Reg ptr y;
 {
    /*** return -iy ***/
  if(REALP(y))
    return complex(ZERO,minus(y));
  if(COMPLEXP(y))
    return complex(CDR(y),minus(CAR(y)));
  error(E_notnum,y);
 }
 
ptr ATAN(py,px)
 Reg ptr py,px;
{
 Reg double y,x;
 y=get_float(py);
 x=get_float(px);
      if(x==0.0)
       {
         if(y==0.0)  error(E_divzero,y);
         if(y<0.0) return minus(ATOMVAL(a_pi2));
         return ATOMVAL(a_pi2);
       }
  y/=x;
  if(x>0.0) return MAKEDBL(atan(y));
  if(y<=0.0) return MAKEDBL(atan(y)+pi);
  return MAKEDBL(atan(y)-pi);
 }

ptr LOG(y)
    Reg ptr y;
 {
   Reg double x;
  if(zerop(y)) error(E_illegal);
   if(REALP(y))
      {
          x=get_float(y);
         if(x<0.0)
           {
             return complex(MAKEDBL(log(-x)),ATOMVAL(a_pi));
           }
         return MAKEDBL(log(x));
       }
    if(COMPLEXP(y))
        {
           x=0.5*log(get_float(norm2(y)));
           y=ATAN(CDR(y),CAR(y));
           psh(y);
           y=complex(MAKEDBL(x),y);
           pull();
           return y;
        }
  error(E_notnum,y);
 }

//  2018.3.14  this is OK!  (why???)
ptr EXP(y)
    Reg ptr y;
 {
   Reg double x,b;
    Reg ptr z;
   if(REALP(y))
      {
   // fprintf(stderr,"EXPDebug real y\n");
          x=get_float(y);
        return MAKEDBL(exp(x));
       }
    if(COMPLEXP(y))
        {
           psh(y);
           x=exp(get_float(CAR(y)));
           z=MAKEDBL(x);
           psh(z);
           //b=get_float(CDR(y));
           y=CIS(CDR(y));
// y=MAKEDBL(x*cos(b));
           //psh(y);
           y=genmul(z,y);
//MAKEDBL(x*sin(b));
           //y=complex(y,z);
           stack-=2;
           return y;
        }
  error(E_notnum,y);
 }

//  This cause Seg. fault  why???

ptr EXP_NOGOOD(y)
    Reg ptr y;
 {
   Reg double x,b;
    Reg ptr z;
   if(REALP(y))
      {
   // fprintf(stderr,"EXPDebug real y\n");
          x=get_float(y);
        return MAKEDBL(exp(x));
       }
    if(COMPLEXP(y))
        {
           psh(y);
           x=exp(get_float(CAR(y)));
           z=MAKEDBL(x);
           if(zerop(z)){
             pull();
             return CIS(CDR(y));
           }
           b=get_float(CDR(y));
           y=MAKEDBL(x*cos(b));
           psh(y);
           z=MAKEDBL(x*sin(b));
           stack-=2;
           if(zerop(z)){
               return y;
            }  
            return complex(y,z);
        }
  error(E_notnum,y);
 }


ptr CIS(y)    /***  cos y+ i sin y  **/
    Reg ptr y;
 {
   Reg double x;
   if(REALP(y))
      {
          x=get_float(y);
          psh(MAKEDBL(cos(x)));
          y=MAKEDBL(sin(x));
        return complex(*pull(),y);
       }
  error(E_notreal,y);
 }

ptr INTEXPT(a,y)
    Reg ptr a,y;    
 {
   /***   y ispos integer, a is any  ****/
    Reg ptr k,x;    
   if(y==ZERO) return ONE;
   if(y==ONE) return a;
   else if(y==TWO) return genmul(a,a);
   else
    {
     psh(k=fixdiv(y,TWO));   /**** 切り捨て整数。not rational!!! ***/
     psh(x=INTEXPT(a,k));
     *(stack-1)=genmul(x,x);
     if(ODDP(y)) *(stack-1)=genmul(*(stack-1),a);
     y=*pull();
     pull();
     return y;
   }
  }

ptr num_coerce(x,a)
  Reg ptr x,a;
 {
   if(INTEGERP(a)) return x;
   if(RATIOP(a)) return x;
   if(DOUBLEP(a)) return MAKEDBL(get_float(x));
   if(COMPLEXP(a)) 
   return complex(MAKEDBL(get_float(x)),DBL0);
   //return complex(MAKEDBL(get_float(x)),MAKEDBL(0.0));
  error(E_notnum,a);
 }
   
     
ptr EXPT(a,y)
    Reg ptr a,y;
 {
   Reg double x;
   if(zerop(a)){
      /** printf("expt a== 0\n"); ***/
      if(zerop(y)) return num_coerce(ONE,a);
      return num_coerce(ZERO,a);
    }
     
   if(INTEGERP(y))
      {
        if(y==ZERO) 
               return  num_coerce(ONE,a);
            psh(y);
            psh(a);
             if(gsign(y)) 
              {
                *(stack-2)=y=minus(y);
                *(stack-1)=a=gendiv(ONE,a);
              }
             y=INTEXPT(a,y);
             sp_pull(2);
             return y;
     }
    psh(a=LOG(a));
    *(stack-1)=a=genmul(y,a);  
    a=EXP(a);             
    pull();
    return a;
 }

ptr s_exp(int narg,ptr env)
    
 {
  checknarg(narg-1);
    return EXP(*(stack-1));
 }

ptr s_expt(int narg,ptr env)
    
 {
  checknarg(narg-2);
    return EXPT(*(stack-2),*(stack-1));
 }

ptr s_cis(int narg,ptr env)
    
 {
  checknarg(narg-1);
    return CIS(*(stack-1));
 }
 
ptr s_atan(int narg,ptr env)
    
 {
   Reg ptr y,x;
   y=*(stack-narg);
   if(narg==2)
    {
      x=*(stack-1);
      if(x==ZERO||(get_float(x)==0.0))
       {
         if(y==ZERO||(get_float(y)==0.0))  error(E_divzero,y);
         if(gsign(y)) return minus(ATOMVAL(a_pi2));
         return ATOMVAL(a_pi2);
       }
      return ATAN(y,x);
    }
   checknarg(narg-1);
   if(REALP(y))
       return MAKEDBL(atan(get_float(y)));
   if(COMPLEXP(y))
    {
      x=complexadd(y,ATOMVAL(a_imag));
      psh(x);
      y=complexsub(ATOMVAL(a_imag),y);
      psh(y);
      *(stack-1)=y=gendiv(x,y);
      *(stack-1)=y=LOG(y);
      *(stack-1)=y=mul_imag(y);
      y=gendiv(y,TWO);
      sp_pull(2);
      return y;
    }
  error(E_notnum,y);
 }

ptr s_sqrt(int narg,ptr env)
    
 {
   Reg double x;
   Reg ptr y=*(stack-1);
   
  checknarg(narg-1);
   if(REALP(y))
     {
         if(gsign(y)) 
           return complex(ZERO,MAKEDBL(sqrt(-get_float(y))));
         return MAKEDBL(sqrt(get_float(y)));
     }
   if(COMPLEXP(y))
    {
       psh(y=LOG(y));
       *(stack-1)=y=gendiv(y,TWO);
       y=EXP(y);
       pull();
       return y;
    } 
  error(E_notnum,y);
 }

ptr s_log(int narg,ptr env)
    
 {
   Reg ptr y,x;
    if(narg==1)
       return y=LOG(*(stack-1));
    checknarg(narg-2);
    
   x=*(stack-1);
      /*** x is base **/
       y=LOG(*(stack-2));
      psh(y);
      if(x==TWO) 
          x=ATOMVAL(a_log2);
      else if(x==TEN) 
          x=ATOMVAL(a_log10);
      else
          x=LOG(x);
      psh(x);
      y=gendiv(y,x);
      sp_pull(2);
    return y;
 }

ptr s_log10(int narg,ptr env)
    
 {
   Reg ptr y,x;
  checknarg(narg-1);
      y=LOG(*(stack-1));
      psh(y);
          x=ATOMVAL(a_log10);
      y=gendiv(y,x);
      pull();
   return y;
 }

ptr s_log2(int narg,ptr env)
    
 {
   Reg ptr y,x;
  checknarg(narg-1);
      y=LOG(*(stack-1));
      psh(y);
          x=ATOMVAL(a_log2);
      y=gendiv(y,x);
      pull();
   return y;
 }  


ptr s_sin(int narg,ptr env)
    
 {
   Reg ptr y=*(stack-1);
  checknarg(narg-1);
   if(REALP(y))
     {
    return MAKEDBL(sin(get_float(y)));
     }
   if(COMPLEXP(y))
    {
      psh(y=mul_imag(y));
      *(stack-1)=y=SINH(y);
      y=mul_imag2(y);
      pull();
     return y;
    }
  error(E_notnum,y);
 }
ptr s_cos(int narg,ptr env)
    
 {
   Reg ptr y=*(stack-1);
  checknarg(narg-1);
   if(REALP(y))
     {
    return MAKEDBL(cos(get_float(y)));
     }
   if(COMPLEXP(y))
    {
      psh(y=mul_imag(y));
      y=COSH(y);
      pull();
     return y;
    }
  error(E_notnum,y);
 }
ptr s_tan(int narg,ptr env)
    
 {
   Reg ptr y=*(stack-1);
  checknarg(narg-1);
   if(REALP(y))
     {
    return MAKEDBL(tan(get_float(y)));
     }
   if(COMPLEXP(y))
    {
      psh(y=mul_imag(y));
      *(stack-1)=y=TANH(y);
      y=mul_imag2(y);
      pull();
     return y;
    }
  error(E_notnum,y);
 }
ptr s_asin(int narg,ptr env)
    
 {
   Reg ptr y=*(stack-1);
   Reg double x;
  checknarg(narg-1);
   if(REALP(y))
     {
        x=get_float(y);
       if(fabs(x)<=1.0)
         return MAKEDBL(asin(x));
       x+=sqrt(x*x-1.0);
       return complex(ATOMVAL(a_pi2),MAKEDBL(-log(x)));
     }
   if(COMPLEXP(y))
    {
      psh(y=mul_imag(y));
      *(stack-1)=y=ASINH(y);
      y=mul_imag2(y);
      pull();
     return y;
    }
  error(E_notnum,y);
 }

ptr s_acos(int narg,ptr env)
    
 {
   Reg ptr y=*(stack-1);
   Reg double x;
  checknarg(narg-1);
   if(REALP(y))
     {
        x=get_float(y);
       if(fabs(x)<=1.0)
         return MAKEDBL(acos(x));
       x+=sqrt(x*x-1.0);
       return complex(ZERO,MAKEDBL(log(x)));
     }
   if(COMPLEXP(y))
    {
      psh(y=ACOSH(y));
      y=mul_imag2(y);
      pull();
     return y;
    }
  error(E_notnum,y);
 }

double ss_sinh(x)
  Reg double x;
 {
  Reg double s,g,p;
  Reg int n;
   if(fabs(x)>0.3) return 0.5*(exp(x)-exp(-x));
   p=x*x;
   g=s=1.0;
   for(n=3;n<16;n+=2)
    {
      g*=p/n/(n-1);
      s+=g;
     }
   return s*x;
  }
double ss_cosh(x)
  Reg double x;
 {
    return 0.5*(exp(x)+exp(-x));
  }
ptr SINH(z)
   Reg ptr z;
 {
   Reg double x,y;
   if(REALP(z))
     {
        x=get_float(z);
         return MAKEDBL(ss_sinh(x));
     }
   if(COMPLEXP(z))
    {
       /***  cos y sinh x+i sin y cosh x  ***/
        x=get_float(CAR(z));
        y=get_float(CDR(z));
        psh(MAKEDBL(cos(y)*ss_sinh(x)));
        z=MAKEDBL(sin(y)*ss_cosh(x));
       return complex(*pull(),z);
     }
  error(E_notnum,y);
 }
ptr COSH(z)
   Reg ptr z;
 {
   Reg double x,y;
   if(REALP(z))
     {
        x=get_float(z);
         return MAKEDBL(ss_cosh(x));
     }
   if(COMPLEXP(z))
    {
       /***  cos y sinh x+i sin y cosh x  ***/
        x=get_float(CAR(z));
        y=get_float(CDR(z));
        psh(MAKEDBL(cos(y)*ss_cosh(x)));
        z=MAKEDBL(sin(y)*ss_sinh(x));
       return complex(*pull(),z);
     }
  error(E_notnum,y);
 }
       
ptr TANH(z)
   Reg ptr z;
 {
   psh(SINH(z));
   psh(z=COSH(z));
   z=gendiv(*(stack-2),z);
   sp_pull(2);
   return z;
 }
ptr ASINH(y)
   Reg ptr y;
 {
   Reg double x;
   Reg ptr *z;
   if(REALP(y))
     {
        x=get_float(y);
       x+=sqrt(x*x+1.0);
       return MAKEDBL(log(x));
     }
   if(COMPLEXP(y))
    {
      return LOGRT(y,ONE);
    }
  error(E_notnum,y);
 }
ptr ACOSH(y)
   Reg ptr y;
 {
   Reg double x;
   Reg ptr *z;
   if(REALP(y))
     {
        x=get_float(y);
       if(fabs(x)>=1.0)
         {
           x+=sqrt(x*x-1.0);
           return MAKEDBL(log(x));
         }
      return LOGRT(y,-1);
     }
   if(COMPLEXP(y))
    {
      return LOGRT(y,-1);
    }
  error(E_notnum,y);
 }

ptr LOGRT(y,c)
   Reg ptr y,c;
 {
   Reg ptr *z;
      z=l_sp;
      psh(genmul(y,y));
      *z=genadd(*z,c);
      *z=LOG(*z);
      *z=gendiv(*z,TWO);
      *z=EXP(*z);
      *z=genadd(y,*z);
      y=LOG(*z);
      pull();
     return y;
   }
ptr ATANH(y)
    Reg ptr y;
 {
   Reg double x;
   Reg ptr *z,w;
   if(NUMBERP(y))
     {
       if(REALP(y))
       {
        x=get_float(y);
       if(fabs(x)<=1.0)
         {
           x=(1.0+x)/(1.0-x);
           return MAKEDBL(0.5*log(x)); 
         }
        }
      psh(w=genadd(y,ONE));
      z=l_sp;
      y=gensub(ONE,y);
      psh(y);
      *z=gendiv(w,y);
      *z=LOG(*z);
      y=gendiv(*z,TWO);
      sp_pull(2);
      return y;
    }
  error(E_notnum,y);
 }

ptr s_sinh(int narg,ptr env)
    
 {
    return SINH(*(stack-1));
 }
ptr s_cosh(int narg,ptr env)
    
 {
    return COSH(*(stack-1));
 }
ptr s_tanh(int narg,ptr env)
    
 {
    return TANH(*(stack-1));
 }
ptr s_asinh(int narg,ptr env)
    
 {
    return ASINH(*(stack-1));
 }
ptr s_acosh(int narg,ptr env)
    
 {
    return ACOSH(*(stack-1));
 }
ptr s_atanh(int narg,ptr env)
    
 {
    return ATANH(*(stack-1));
 }

 ptr s_isqrt(int narg,ptr env)
  {
    Reg ptr x,n,c,y,*sp;
    Reg double v;
    x=*(stack-1);
    if(SINTP(x)){
        v=get_float(x);
        v=sqrt(v);
        return MAKESINT((int)floor(v));
     }
    else if(BIGINTP(x)){
        mpz_sqrt(mpzwork1,BIGNUM(x));
        return mpz_check16(mpzwork1);
     }

    error(E_notint,x);
}
 


