/***   list_circle.c   ****/
#include "lisp.h"
#include <stdio.h>
/****  Circular list   ***/
/****   05.10.8   separated from list.c  ***/


extern ptr clear_circle(); 

static int dot_list, circle_length;
static ptr list_equal_circular();

void clear_circle_list(Reg ptr q)
{
     if(!CIRCLEP(q)) return;
      CLRCIRCLE(q);
      while(1)
       {
        q=CDR(q);
         if(q==NIL) break;
         if(!SREGP(q)&&CONSP(q))  /***  q maybe integer ***/
          {
                if(!CIRCLEP(q)){
                    break;
                }
             CLRCIRCLE(q);
           }
          else
           {  /***  dot   pair  ***/
            break;
           }
      }
}

/**** return -1 for circular list, dot_list is set   ***/
int list_length(Reg ptr q) 
{   
  ptr memo=q,car;
  int n=0;
      dot_list=0;
      if(q==NIL) return 0;
      if(SREGP(q)||!CONSP(q)) error(E_notlist,q);
      SETCIRCLE(q);
      while(1)
       {
           q=CDR(q);
           n++;
         if(q==NIL) break;
         if(!SREGP(q)&&CONSP(q))  /***  q maybe integer ***/
          {
                if(CIRCLEP(q)){
                    circle_length=n;
                    n=-1;
                    break;
                }
             SETCIRCLE(q);
           }
          else
           {  /***  dot   pair  ***/
             dot_list=YES;
            break;
           }
        }
    clear_circle_list(memo);
 return n;
}

ptr s_list_length(int narg,ptr env)
      /* (length L)   error for improper or circular */
 {
    Reg int s=0;
    Reg ptr q;
    /** see CL p.205  これは, 循環list ならnil を返す。->a_f for ss ***/
    /***  write.cのアルゴリズムでやる　  ***/
                                                                                            
    checknarg(--narg);
    q=*(stack-1);
    s=list_length(q);
    if((s<0)||dot_list) error(E_notlist,q);
    return MAKENUM(s);
}

ptr s_list_length_plus(int narg,ptr env)
      /* (length L)   error for improper r */
 {
    Reg int s=0;
    Reg ptr q;
    /** see CL p.205  SRFI-1 これは, 循環list ならnil を返す。->a_f for ss ***/
    checknarg(--narg);
    q=*(stack-1);
    s=list_length(q);
    if(s<0) return a_f;
    if(dot_list) error(E_notlist,q);
    return MAKENUM(s);
}

ptr s_list_length_pp(int narg,ptr env)
      /* (length L)   error for improper r */
 {
    Reg int s=0;
    Reg ptr q;
    /** これは, 循環list なら circle_length を返す ***/
    checknarg(--narg);
    q=*(stack-1);
    s=list_length(q);
    if(s<0) s=circle_length;
    return MAKENUM(s);
}

ptr s_circular_list(int narg,ptr env)
      /* (length L)   error for improper r */
 {
    Reg int s=0;
    Reg ptr q;
    /** see CL p.205  SRFI-1 これは, 循環list ならnil を返す。->a_f for ss ***/
    checknarg(--narg);
    q=*(stack-1);
   if(SREGP(q)||!CONSP(q)) return a_f;
    s=list_length(q);
    if(s<0) return a_t;
    return a_f;
}

ptr s_dotted_list(int narg,ptr env)
      /* (length L)   error for improper r */
 {
    Reg int s=0;
    Reg ptr q;
    /** see CL p.205  SRFI-1 これは, 循環list ならnil を返す。->a_f for ss ***/
    checknarg(--narg);
    q=*(stack-1);
   if(q==a_nil) return a_f;
   if(SREGP(q)||!CONSP(q)) return a_t; /*** as dotted list of length 0! ***/
    s=list_length(q);
    if(s<0) return a_f;
    return PRED(dot_list);
}

/****   simple list equal, same length  ***/ 
ptr list_equal_simple(Reg ptr p, Reg ptr q,ptr func){
  ptr *base=stack,s=a_t;
  stack+=2;
  while(CONSP(p)){
     *base=CAR(p);
     *(base+1)=CAR(q);
     s=funcall(func,2,a_nil);
     if(s==a_f){
         goto out;
     }
     p=CDR(p);
     q=CDR(q);
  }
out:
 stack-=2;
 return s;
}

/****   list equal include circular  ***/ 
static ptr list_equal(Reg ptr p, Reg ptr q,ptr func)
{
  ptr memo=q,car;
  int n,m;
    if(p==q) return a_t;
    n=list_length(p);  
    if(dot_list) error(E_notlist,p);
    m=list_length(q);  
    if(dot_list) error(E_notlist,q);
    if(n!=m) return a_f;
    if(n<0) return list_equal_circular(p,q,func);    
    return list_equal_simple(p,q,func);   
} 

int check_equal_circular(ptr p,ptr q,ptr r,ptr v){
   if(p==r) return (q==v);
   return  check_equal_circular(p,q,CDR(r),CDR(v));
}

static ptr list_equal_circular(ptr p, ptr q,ptr func){
   ptr *base=stack,r=p,v=q,s=a_t;    
      SETCIRCLE(q);
      SETCIRCLE(p);
  stack+=2;
      while(1)
       {
         *base=CAR(p);
         *(base+1)=CAR(q);
         if(*base!=*(base+1))
             s=funcall(func,2,a_nil);
         if(s==a_f)  goto out;
          p=CDR(p);
          q=CDR(q);
         if(CIRCLEP(p)){
            if(CIRCLEP(q))  break;
            s=a_f;
            goto out;
         }
             SETCIRCLE(p);
             SETCIRCLE(q);
       }
      s=PRED(check_equal_circular(p,q,r,v));
out:
     stack-=2;
    clear_circle_list(r);
    clear_circle_list(v);
 return s;
}

ptr s_list_equal(int narg,ptr env){
  ptr func,*base,p,q,s=a_t;
  if(narg==0) checknarg(narg-1);
  base=stack-narg;
  func=*base++;
  narg--;
  if(narg<2) return a_t;
   p=*base++;
   if(SREGP(p)||!CONSP(p))
     if(p!=a_nil) error(E_notlist,p);
  while(narg>1){
   q=*base++;
   if(SREGP(q)||!CONSP(q))
     if(q!=a_nil) error(E_notlist,q);
   if(p!=q)
      s=list_equal(p,q,func);
   if(s==a_f) break;
   p=q;
   narg--;
  }
  return s;
}
   
  

