/*
 * matrix . c 			-- matrix management
 *  
 * Copyright 2002 Sadao Sato
 *  extension for 2-dim matrix
 *
 */
#define  MATRIX_C

#include "lisp.h"
#include <string.h>


static int ans=0;
static ptr *workptr=0;
static int worksize=0;

extern char *alloc_space();
extern ptr gendiv(),genadd(),genmul(),minus();

ptr matrix_base_s();
ptr s_matrix_det_s();
static  ptr matrix_resize(ptr mat, int new1,int new2, ptr fill);
void matrix_rotate_right();
void matrix_rotate_left();

static ptr *getwork(int n){
  if(n<=worksize) return workptr;
  if(workptr) Myfree(workptr);
  worksize=n;
  return  workptr=(ptr *)Mymalloc(n*sizeof(ptr));
}

static ptr makematrix(int lenr,int lenc, ptr init)
{
  int j;
  ptr  z,*p;
  /******    CDR is CONS(n m)   GBC caution!  n,m are SINT ****/

  z=new_cell(MATRIX);
  psh(z);
  CDR(z)=CONS(MAKESINT(lenr),MAKESINT(lenc));
  j=lenr*lenc;
  p=(ptr *)alloc_space(z, j*sizeof(ptr));  /*** CAR(c) is set ***/
    while(j--) *p++=init;  /*** init ***/
  stack--;
   return z;
  }

void checkmatrix(ptr x){
  if(SREGP(x)||!MATRIXP(x)) error(E_notarray,x);
}

/*** used by write.c  ***/
ptr  *get_matrix(Reg ptr mat,int *len1,int *len2){
 ptr p;
    p=CDR(mat);
  
   *len1=SINT_NUMB(CAR(p));
   *len2=SINT_NUMB(CDR(p));
    p=CAR(mat);
   return(ptr *)(++p);
}

ptr s_matrixp(int narg,ptr env){
   ptr x ;
    x=*(stack-1);
  if(SREGP(x)||!MATRIXP(x)) return a_f;
  return a_t;
}

ptr s_make_matrix(int narg,ptr env){
   ptr s1,s2,ini ;
   if(narg==2) ini=a_undef;
   else if(narg==3) ini=*(stack-1);
   else
       checknarg(narg-3);
    s1=*(stack-narg);
    s2=*(stack-narg+1);
   return makematrix(SINT_NUMB(s1),SINT_NUMB(s2),ini);
}

ptr s_matrix_ref(int narg,ptr env){
  Reg  ptr mat,*p ;
  int n,m,s1,s2;
    mat=*(stack-3);
   n= SINT_NUMB((long)*(stack-2));
   m= SINT_NUMB((long)*(stack-1));
   checknarg(narg-3);
   checkmatrix(mat);
   p=get_matrix(mat,&s1,&s2);
  /***
   printf("n %x m %x \n",*(stack-2),*(stack-1));
   printf("n %d m %d \n",n,m);
   printf("s1 %d s2 %d \n",s1,s2);
  ***/
   if(n>=s1)  error(E_range,*(stack-2));
   if(m>=s2)  error(E_range,*(stack-1));
   return *(p+n*s2+m);
}

ptr s_matrix_set(int narg,ptr env){
   Reg ptr mat,*p ,val;
  int n,m,s1,s2;
    mat=*(stack-4);
   n= SINT_NUMB(*(stack-3));
   m= SINT_NUMB(*(stack-2));
    val=*(stack-1);
   checknarg(narg-4);
   checkmatrix(mat);
   p=get_matrix(mat,&s1,&s2);
   if(n>=s1)  error(E_range,*(stack-3));
   if(m>=s2)  error(E_range,*(stack-2));
    *(p+n*s2+m)=val;
  return a_undef;
}

ptr make_matrix_list(ptr x){  /****  list->matrix ***/
  Reg int m2;
  Reg ptr y,*p,v;
  int n,m;
   n=_length(x);
   m2=m=_length(CAR(x));
   v=makematrix(n,m,a_nil);
   p=get_matrix(v,&n,&m);
  while(n--){
    y=CAR(x);
   m=m2;
    while(m--){
        *p++=CAR(y);
        y=CDR(y);
    }
   x=CDR(x);
  }
  return v;
}

ptr s_matrix(int narg,ptr env){  /****  list1 list2  ->matrix ***/
  Reg int m2;
  Reg ptr y,*p,v,*sfree;
  int n,m;

  sfree=stack-narg;
   n=narg;
   m2=m=_length(*sfree);
   v=makematrix(n,m,a_nil);
   p=get_matrix(v,&n,&m);
  while(n--){
    y=*sfree++;
   m=m2;
    while(m--){
        *p++=CAR(y);
        y=CDR(y);
    }
  }
  return v;
}


ptr s_list2matrix(int narg,ptr env)
      /* (list->vector L) */
{
    Reg int s=0;
    checknarg(narg-1);
   return make_matrix_list(*(stack-1));
}


static ptr matrix_2list(ptr v){  /**** matrix -> list ***/
  int n,m,m2;
  Reg ptr x=a_nil,y=a_nil,*p;
   checkmatrix(v);
   p=get_matrix(v,&n,&m);
  p+=n*m;
  m2=m;
  psh(x);
  while(n--){
    m=m2;
    x=a_nil;
    while(m--){
       x=SCONS(*--p,x);
    }
   *(stack-1)=y=SCONS(x,y);
  }
  return *pull();
}


ptr s_matrix2list(int narg,ptr env)
      /* (mat->list) */
{
    checknarg(narg-1);
   return matrix_2list(*(stack-1));
}

ptr s_matrix_length(Reg int narg,ptr env)
      /* (vector x1 x2 ....) */
{
   Reg ptr x;
    int n;
    checknarg(narg-1);
    x=*(stack-1);
   checkmatrix(x);
    /** get_vector(x,&n); ***/
    return CDR(x);
}


ptr  s_vector2matrix(int narg, ptr env)
{
  ptr  x=*(stack-1),z;
  int n,m;
  char *v,*q;
   checknarg(narg-1);

  checkvect(x);
   v=(char *)get_vector(x,&n);

  z=makematrix(n,1,a_nil);
  q=(char *)get_matrix(z,&n,&m);
   memcpy(q, v, n * sizeof(ptr));
  return z;
}

ptr s_matrix_fill(Reg int narg,ptr env)
      /*  */
{
   Reg ptr x,ini,*p;
    int n,m;
    checknarg(narg-2);
    x=*(stack-2);
    ini=*(stack-1);
   checkmatrix(x);
    p=get_matrix(x,&n,&m);
   n*=m;
    while(n--){
    *p++=ini;
    }
    return a_undef;
}


/*
 * ss  bonus
 */
ptr s_matrix_map_1(int narg,ptr env)
{
  ptr x,fct,gyo;
  int n,m,k;
  ptr *p;
   checknarg(narg-3);
    x=*(stack-3);
    fct=*(stack-2);
    gyo=*(stack-1);
   checkmatrix(x);
    p=get_matrix(x,&n,&m);
   k=SINT_NUMB(gyo);
  if(k>=n)  error(E_range, gyo);
   p+=m*k;
   psh(x);
   for(n=0;n<m;n++){
     *(stack-1)=*p;
     *p++=funcall(fct,1,a_nil);
     }
  pull();
  return x;
}

ptr s_matrix_t_map_1(int narg,ptr env)
{
  ptr x,fct,retu;
  int n,m,k;
  ptr *p;
   checknarg(narg-3);
    x=*(stack-3);
    fct=*(stack-2);
    retu=*(stack-1);
   checkmatrix(x);
    p=get_matrix(x,&n,&m);
   k=SINT_NUMB(retu);
  if(k>=n)  error(E_range, retu);
   p+=k;
   psh(x);
   for(n=0;n<m;n++){
     *(stack-1)=*p;
     *p=funcall(fct,1,a_nil);
      p+=m;
     }
   pull();
  return x;
}

ptr s_matrix_map_2(int narg,ptr env)
{
  ptr x,fct,gyo1,gyo2;
  int n,m,k,dest;
  ptr *p,*q;
   checknarg(narg-4);
    x=*(stack-4);
    fct=*(stack-3);
    gyo1=*(stack-2);
    gyo2=*(stack-1);
   checkmatrix(x);
    q=p=get_matrix(x,&n,&m);
   k=SINT_NUMB(gyo1);
   dest=SINT_NUMB(gyo2);
  if(k>=n)  error(E_range, gyo1);
  if(dest>=n)  error(E_range, gyo2);
   p+=m*k;
   q+=m*dest;
   stack+=2;
   for(n=0;n<m;n++){
     *(stack-2)=*p++;
     *(stack-1)=*q;
     *q++=funcall(fct,2,a_nil);
     }
   stack-=2;
  return x;
}


ptr s_matrix_t_map_2(int narg,ptr env)
{
  ptr x,fct,retu1,retu2;
  int n,m,k,dest;
  ptr *p,*q;
   checknarg(narg-4);
    x=*(stack-4);
    fct=*(stack-3);
    retu1=*(stack-2);
    retu2=*(stack-1);
   checkmatrix(x);
    q=p=get_matrix(x,&n,&m);
   k=SINT_NUMB(retu1);
   dest=SINT_NUMB(retu2);
  if(k>=m)  error(E_range, retu1);
  if(dest>=m)  error(E_range, retu2);
   p+=k;
   q+=dest;
   stack+=2;
   for(n=0;n<m;n++){
     *(stack-2)=*p;
     *(stack-1)=*q;
     *q=funcall(fct,2,a_nil);
     p+=m;
     q+=m;
     }
   stack-=2;
  return x;
}

ptr s_matrix_base_1(int narg,ptr env)
{
  ptr x,val,gyo;
  int n,m,k;
  ptr *p;
   checknarg(narg-3);
    x=*(stack-3);
    val=*(stack-2);
    gyo=*(stack-1);
   checkmatrix(x);
    p=get_matrix(x,&n,&m);
   k=SINT_NUMB(gyo);
  if(k>=n)  error(E_range, gyo);
   p+=m*k;
   for(n=0;n<m;n++){
     /*** *p++=genmul(*p,val);   compiler bug?  ***/
     *p=genmul(*p,val);
      p++;
     }
  return x;
}

ptr s_matrix_t_base_1(int narg,ptr env)
{
  ptr x,val,retu;
  int n,m,k;
  ptr *p;
   checknarg(narg-3);
    x=*(stack-3);
    val=*(stack-2);
    retu=*(stack-1);
   checkmatrix(x);
    p=get_matrix(x,&n,&m);
   k=SINT_NUMB(retu);
  if(k>=n)  error(E_range, retu);
   p+=k;
   psh(x);
   for(n=0;n<m;n++){
     *p=genmul(*p,val);
      p+=m;
     }
   pull();
  return x;
}

ptr s_matrix_base_2(int narg,ptr env)
{
  ptr x,val,gyo1,gyo2;
  int n,m,k,dest;
  ptr *p,*q,s;
   checknarg(narg-4);
    x=*(stack-4);
    val=*(stack-3);
    gyo1=*(stack-2);
    gyo2=*(stack-1);
   checkmatrix(x);
    q=p=get_matrix(x,&n,&m);
   k=SINT_NUMB(gyo1);
   dest=SINT_NUMB(gyo2);
  if(k>=n)  error(E_range, gyo1);
  if(dest>=n)  error(E_range, gyo2);
   p+=m*k;
   q+=m*dest;
   stack++;
   for(n=0;n<m;n++){
     s=*(stack-1)=genmul(*p,val);
     p++;
     *q=genadd(*q,s);
     q++;
     }
   stack--;
  return x;
}



ptr s_matrix_t_base_2(int narg,ptr env)
{
  ptr x,val,retu1,retu2;
  int n,m,k,dest;
  ptr *p,*q,s;
   checknarg(narg-4);
    x=*(stack-4);
    val=*(stack-3);
    retu1=*(stack-2);
    retu2=*(stack-1);
   checkmatrix(x);
    q=p=get_matrix(x,&n,&m);
   k=SINT_NUMB(retu1);
   dest=SINT_NUMB(retu2);
  if(k>=m)  error(E_range, retu1);
  if(dest>=m)  error(E_range, retu2);
   p+=k;
   q+=dest;
   stack++;
   for(n=0;n<m;n++){
     s=*(stack-1)=genmul(*p,val);
     *q=genadd(*q,s);
     p+=m;
     q+=m;
     }
   stack--;
  return x;
}

ptr s_matrix_swap(int narg,ptr env)
{
  ptr x,gyo1,gyo2;
  int n,m,k,dest;
  ptr *p,*q,s;
   checknarg(narg-3);
    x=*(stack-3);
    gyo1=*(stack-2);
    gyo2=*(stack-1);
   checkmatrix(x);
    q=p=get_matrix(x,&n,&m);
   k=SINT_NUMB(gyo1);
   dest=SINT_NUMB(gyo2);
  if(k>=n)  error(E_range, gyo1);
  if(dest>=n)  error(E_range, gyo2);
   p+=m*k;
   q+=m*dest;
   for(n=0;n<m;n++){
     s=*p;
     *p++=*q;
    *q++=s;
     }
  return x;
}

ptr s_matrix_t_swap(int narg,ptr env)
{
  ptr x,retu1,retu2;
  int n,m,k,dest;
  ptr *p,*q,s;
   checknarg(narg-3);
    x=*(stack-3);
    retu1=*(stack-2);
    retu2=*(stack-1);
   checkmatrix(x);
    q=p=get_matrix(x,&n,&m);
   k=SINT_NUMB(retu1);
   dest=SINT_NUMB(retu2);
  if(k>=m)  error(E_range, retu1);
  if(dest>=m)  error(E_range, retu2);
   p+=k;
   q+=dest;
   for(n=0;n<m;n++){
     s=*p;
     *p=*q;
     *q=s;
     p+=m;
     q+=m;
     }
  return x;
}


ptr s_matrix_copy(int narg,ptr env)
{
  ptr  v,z,*p,*q;
  int n,m;
   checknarg(narg-1);
    v=*(stack-1);
   checkmatrix(v);
    p=get_matrix(v,&n,&m);

  z = makematrix(n,m, a_nil);
    p=get_matrix(v,&n,&m);
    q=get_matrix(z,&n,&m);
  memcpy(q, p, n *m* sizeof(ptr));
  return z;
}

ptr s_matrix_transpose(int narg,ptr env)
{
  ptr  v,z,*p,*q,*r;
  int n,m,i,j,skip;
   checknarg(narg-1);
    v=*(stack-1);
   checkmatrix(v);
    p=get_matrix(v,&n,&m);

  z = makematrix(m,n, a_nil);
    p=get_matrix(v,&n,&m);
    q=get_matrix(z,&m,&n);
 if((n!=1)&&(m!=1)){
  skip=n;
    for(i=0;i<n;i++){
     r=q++;
     for(j=0;j<m;j++){
       *r=*p++;
        r+=skip;
     }
    }
 }
else
  memcpy(q, p, n *m* sizeof(ptr));
  return z;
}

ptr s_matrix_transpose_s(int narg,ptr env)
{
  ptr  v,z,s,*p,*q,*r;
  int n,m,i,j,skip;
   checknarg(narg-1);
    v=*(stack-1);
   checkmatrix(v);
    p=get_matrix(v,&n,&m);

 if((n==1)||(m==1)) goto sw;
    q=getwork(n*m);
  skip=n;
    for(i=0;i<n;i++){
     r=q++;
     for(j=0;j<m;j++){
       *r=*p++;
        r+=skip;
     }
    }
  p=get_matrix(v,&n,&m);
  memcpy(p, workptr, n*m*sizeof(ptr));
sw:
  z=CDR(v);
  s=CAR(z);
  CAR(z)=CDR(z);
  CDR(z)=s; 
  return v;
}


static  ptr matrix_resize(ptr mat, int new1,int new2, ptr fill)
{
   ptr z;
   int i,j,old1,old2,m1,m2;
   ptr *p,*q,*r,*newp;
   p=get_matrix(mat,&old1,&old2);
  z=makematrix(new1,new2,fill);
  newp=q =get_matrix(z,&new1,&new2);
   p=get_matrix(mat,&old1,&old2);
  m1=new1;
  if(old1<new1) m1=old1;
  m2=new2;
  if(old2<new2) m2=old2;

  for(i=0;i<m1;i++){
    r=p;
    p+=old2;
    q=newp;
    newp+=new2;
   for(j=0;j<m2;j++){
         *q++=*r++;
   }
  }
  return z;
}

ptr s_matrix_resize(int narg,ptr env)
{ 
  ptr mat, size,size2, fill;
 
  int new1,new2;
  checknarg(narg-4);
   mat=*(stack-4);
    size=*(stack-3);
    size2=*(stack-2);
    fill=*(stack-1);
   checkmatrix(mat);
  if ((new1=SINT_NUMB(size))<0) error(E_range,size);
  if ((new2=SINT_NUMB(size2))<0) error(E_range,size2);
 return matrix_resize(mat,new1,new2,fill);
}


ptr s_matrix_mul_set(int narg,ptr env)
{
   ptr  mat1,  mat2, mat3;
  ptr s,m,z,g;
  int i,j,k,r1,c1,r2,c2,r3,c3;
   ptr *p,*q,*r,*tp,*tq,*tp2,*tq2;
  checknarg(narg-3);
   mat1=*(stack-3);
   mat2=*(stack-2);
   mat3=*(stack-1);
   checkmatrix(mat1);
   checkmatrix(mat2);
   checkmatrix(mat3);

   p=get_matrix(mat1,&r1,&c1);
   q=get_matrix(mat2,&r2,&c2);
   r=get_matrix(mat3,&r3,&c3);
  if((c1!=r2)||(r1!=r3)||(c2!=c3))  error(E_range, mat3);
  stack+=2;
  for(i=0;i<r1;i++){
   tp=p;
   tq=q;
   p+=c1;
   for(j=0;j<c2;j++){
     s=ZERO;
     tq2=tq++;
     tp2=tp;
     for(k=0;k<c1;k++){
       if((!zerop(m=*tp2++))&&(!zerop(z=*tq2))){
          *(stack-1)=m=genmul(m,z);
          *(stack-2)=s=genadd(s,m);
       }
      tq2+=c2;
     }
    *r++=s;
   }
  }
  stack-=2;
  return mat3;
}

ptr s_matrix_mul(int narg,ptr env)
{
   ptr  mat1,  mat2, mat3;
  ptr s,m,z,g;
  int i,j,k,r1,c1,r2,c2,r3,c3;
   ptr *p,*q,*r,*tp,*tq,*tp2,*tq2;
  checknarg(narg-2);
   mat1=*(stack-2);
   mat2=*(stack-1);
   checkmatrix(mat1);
   checkmatrix(mat2);

   p=get_matrix(mat1,&r1,&c1);
   q=get_matrix(mat2,&r2,&c2);
   r3=r1;
  c3=c2;
  if(c1!=r2)  error(E_range, mat2);
   mat3 = makematrix(r3,c3, a_nil);
   psh(mat3);
   s_matrix_mul_set(3,env);
   pull();
  return mat3;
}

ptr s_matrix_mul_tr_set(int narg,ptr env)
{
   ptr  mat1,  mat2, mat3;
  ptr s,m,z,g,tmp,tmp2;
  int i,j,k,r1,c1,r2,c2,r3,c3;
   ptr *p,*q,*r,*tp,*tq,*tp2,*tq2;
  checknarg(narg-3);
   mat1=*(stack-3);
   mat2=*(stack-2);
   mat3=*(stack-1);
   checkmatrix(mat1);
   checkmatrix(mat2);
   checkmatrix(mat3);

   p=get_matrix(mat1,&r1,&c1);
   q=get_matrix(mat2,&r2,&c2);
   r=get_matrix(mat3,&r3,&c3);

  if((c1!=c2)||(r1!=r3)||(c3!=r2))  error(E_range, mat2);
  stack+=2;
  for(i=0;i<r1;i++){
   tp=p;
   p+=c1;
   tq2=q;
   for(j=0;j<c3;j++){
     s=ZERO;
     tp2=tp;
     for(k=0;k<c1;k++){
        if(zerop(tmp=*tp2++))
           tq2++;
       else
         if(!zerop(tmp2=*tq2++)){
             m=tmp;
             *(stack-2)=m=genmul(m,tmp2);
             *(stack-1)=s=genadd(s,m);
         }
     }
    *r++=s;
   }
  }
  stack-=2;
  return mat3;
}

ptr s_matrix_mul_tr(int narg,ptr env)
{
   ptr  mat1,  mat2, mat3;
  ptr s,m,z,g;
  int i,j,k,r1,c1,r2,c2,r3,c3;
   ptr *p,*q,*r,*tp,*tq,*tp2,*tq2;
  checknarg(narg-2);
   mat1=*(stack-2);
   mat2=*(stack-1);
   checkmatrix(mat1);
   checkmatrix(mat2);

   p=get_matrix(mat1,&r1,&c1);
   q=get_matrix(mat2,&r2,&c2);
   r3=r1;
  c3=r2;
  if(c1!=c2)  error(E_range, mat2);
   mat3 = makematrix(r3,c3, a_nil);
   psh(mat3);
   s_matrix_mul_tr_set(3,env);
   pull();
  return mat3;
}


ptr s_matrix_add(int narg,ptr env)
{
  ptr  mat1,  mat2;
  ptr s,m,z,mat3;
  int i,j,k,r1,c1,r2,c2,r3,c3;
   ptr *p,*q,*r;
  checknarg(narg-2);
   mat1=*(stack-2);
   mat2=*(stack-1);
   checkmatrix(mat1);
   checkmatrix(mat2);

   p=get_matrix(mat1,&r1,&c1);
   q=get_matrix(mat2,&r2,&c2);
  r3=r1;
  c3=c1;
  if((r1!=r2)||(c1!=c2))  error(E_range, mat2);
  mat3 = makematrix(r3,c3, a_nil);
  psh(mat3);
   p=get_matrix(mat1,&r1,&c1);
   q=get_matrix(mat2,&r2,&c2);
   r=get_matrix(mat3,&r3,&c3);

  for(i=0;i<r1*c1;i++){
     *r++= genadd(*p++,*q++);
  }
  pull();
  return mat3;
}

ptr s_matrix_add_set(int narg,ptr env)
{
  ptr  mat1,  mat2;
  ptr s,m,z,mat3;
  int i,j,k,r1,c1,r2,c2,r3,c3;
   ptr *p,*q,*r;
  checknarg(narg-3);
   mat1=*(stack-3);
   mat2=*(stack-2);
   mat3=*(stack-1);
   checkmatrix(mat1);
   checkmatrix(mat2);
   checkmatrix(mat3);

   p=get_matrix(mat1,&r1,&c1);
   q=get_matrix(mat2,&r2,&c2);
   r=get_matrix(mat3,&r3,&c3);
  if((r1!=r2)||(c1!=c2))  error(E_range, mat2);
  if((r1!=r3)||(c1!=c3))  error(E_range, mat3);

  for(i=0;i<r1*c1;i++){
     *r++= genadd(*p++,*q++);
  }
  return mat3;
}

ptr find_nonzero(ptr *p,int i,int j,int n,int m)
{
  ptr z;
  int k;
  p+=i*m+j;
  for(k=i;k<n;k++){
   z=*p;
  if(!zerop(z)){
    ans=k;
    return z;
  }
  p+=m;
  }
 return a_nil;
}

void matrix_swap(ptr *p,ptr *q,int k)
{
  ptr  z;
  int i;
  for(i=0;i<k;i++){
   z=*p;
   *p++=*q;
   *q++=z;
  }
}
  
ptr s_matrix_det(int narg, ptr env)
{
  ptr  mat,z;
  checknarg(narg-1);
  checkmatrix(mat=*(stack-1));
  z= s_matrix_copy(narg,env);
  psh(z);
  z=s_matrix_det_s(1,env);
  pull();
  return z;
}

ptr s_matrix_det_s(int narg, ptr env)
{
  ptr  mat,z;
  ptr  det,d,g;
  int n,m,i,j,k,l;
   ptr *base,*p,*q,*r,*p2;
    checknarg(narg-1);
  checkmatrix(mat=*(stack-1));

   det=ONE;
   base=p=q=get_matrix(mat,&n,&m);
   stack+=3;
   for(i=0;i<n;i++){
    z=*p;
    if(zerop(z)){
       z=find_nonzero(base,i+1,i,n,m);
       if(z==a_nil) return ZERO;
       matrix_swap(p,p+(ans-i)*m,m-i);
       det=minus(det);
     }
    *(stack-1)=det;
    *(stack-1)=det=genmul(det,z);
    q=p;
    /*** sweep **/
    for(j=i+1;j<n;j++){
      q+=m;
      r=q;   
         *(stack-2)=d=*r;
        if(!zerop(d)){
           *(stack-2)=d=gendiv(d,z);
           *(stack-2)=d=minus(d);
          p2=p+1;
          *r++=ZERO; /** always 0 **/
          for(l=i+1;l<m;l++){
             g=*p2++;
            *(stack-3)=g=genmul(g,d);
            g=genadd(g,*r); 
            *r++=g; 
          }
        }
    }
   p+=(m+1);
  }
  stack-=3;
 return det;
}

ptr s_matrix_inverse(int narg, ptr env)
{
  ptr  mat,z;
  ptr  det,d,g;
  int n,m,i,j,k,l;
   ptr *base,*p,*q,*r,*p2;
    checknarg(narg-1);
  checkmatrix(mat=*(stack-1));

   p=get_matrix(mat,&n,&m);

  if(n!=m)   error(E_range, mat);
  z=matrix_resize(mat,n,2*n,ZERO);
  psh(z);
   base=p=get_matrix(z,&n,&m);
  p+=n;
  /*  make En */
  for(i=0;i<n;i++){
   *p=ONE;
   p+=n+n+1;
  }
  *(stack-1)=z=matrix_base_s(z);
   base=p=get_matrix(z,&n,&m);
  if(zerop(*(base+n*(2*n-1)-1)))  error(E_divzero, mat);
  p=base;
  q=base+n;
  for(i=0;i<n;i++){
    for(j=0;j<n;j++){
     *p++=*q++;
    }
   p+=n;
   q+=n;
  }
  z=matrix_resize(z,n,n,a_nil);
  pull();
  return z;
}

ptr s_matrix_base_s(int narg,ptr env)
{
  ptr mat, z;
    checknarg(narg-1);                     
  checkmatrix(mat=*(stack-1));

  z= s_matrix_copy(1,env);
  psh(z);
  z=matrix_base_s(z);
  pull();
  return z;
}

ptr matrix_base_s(ptr  mat)
{      
 /****   kihon henkei  **/
 ptr  z,d,g;
  int n,m,i,i2,j,k,l;
   ptr *base,*p,*q,*r,*p2;
  checkmatrix(mat);
   base=p=q=get_matrix(mat,&n,&m);

   i2=0;
   /** i=retu , i2=gyo **/
   stack+=2;
   for(i=0;i<m;i++){
    if(i2==n) break;
    z=*p;
    if(zerop(z)){
       /* printf("zero %d %d \n", i, i2);*/
       z=find_nonzero(base,i2+1,i,n,m);
       if(z==a_nil) {
          p++;
          continue;
       }
       matrix_swap(p,p+(ans-i2)*m,m-i);
     }
    q=p;
   /* printf(" swap done %d %d \n", i, i2); */
    for(j=0;j<m-i;j++){
     g=*q;
     g=gendiv(g,z);
     *q++=g;
    }
   /* printf(" norm done %d %d \n", i, i2); */
    q=base+i;
    /*** sweep **/
    for(j=0;j<n;j++){
      r=q;
      q+=m;
        if(j==i2) continue;
         d=*r;
        if(!zerop(d)){
          *(stack-1)=d=minus(d);
          p2=p+1;
          *r++=ZERO; /** always 0 **/
          for(l=i+1;l<m;l++){
           g=*p2++;
           *(stack-2)=g=genmul(g,d);
           g=genadd(g,*r);
           *r++=g;
          }
        }
    }

   /* printf("sweep done %d %d \n", i, i2); */
   p+=(m+1);
   i2++;
  }
  stack-=2;
 return mat;
}

/***  2003.6.23  **/
ptr s_matrix_append(int narg, ptr env)
{
  ptr  mat1,mat2,mat3,s,m,z;
  int i,j,k,r1,c1,r2,c2,r3,c3;
   ptr *base,*p,*q,*r;
    checknarg(narg-2);
  mat1=*(stack-2);
  mat2=*(stack-1);
  checkmatrix(mat1);
  checkmatrix(mat2);

   p=get_matrix(mat1,&r1,&c1);
   q=get_matrix(mat2,&r2,&c2);

  r3=r1;
  c3=c1+c2;
  if((r1!=r2))  error(E_range, mat2);
  mat3 = makematrix(r3,c3, a_nil);
   p=get_matrix(mat1,&r1,&c1);
   q=get_matrix(mat2,&r2,&c2);
   r=get_matrix(mat3,&r3,&c3);
  psh(mat3);
  for(i=0;i<r1;i++){
    for(j=0;j<c1;j++){
      *r++=*p++;
    }
    for(j=0;j<c2;j++){
      *r++=*q++;
    }
  }
  pull();
  return mat3;
}

ptr s_matrix_append_tr(int narg, ptr env)
{
  ptr  mat1,mat2,mat3,s,m,z;
  int i,j,k,r1,c1,r2,c2,r3,c3;
   ptr *base,*p,*q,*r;
    checknarg(narg-2);
  mat1=*(stack-2);
  mat2=*(stack-1);
  checkmatrix(mat1);
  checkmatrix(mat2);

   p=get_matrix(mat1,&r1,&c1);
   q=get_matrix(mat2,&r2,&c2);

  r3=r1+r2;
  c3=c1;
  if((c1!=c2))  error(E_range, mat2);
  mat3 = makematrix(r3,c3, a_nil);
   p=get_matrix(mat1,&r1,&c1);
   q=get_matrix(mat2,&r2,&c2);
   r=get_matrix(mat3,&r3,&c3);
  psh(mat3);
 for(i=0;i<r1;i++){
    for(j=0;j<c1;j++){
      *r++=*p++;
    }
  }
  for(i=0;i<r2;i++){
    for(j=0;j<c1;j++){
      *r++=*q++;
    }
  }
  pull();
  return mat3;
}

ptr s_matrix_divide(int narg, ptr env)
{
  ptr  mat1,mat2,mat3,gyo,z;
  int i,j,k,r1,c1,r2,c2,r3,c3;
   ptr *p,*q,*r;
    checknarg(narg-2);
  mat1=*(stack-2);
  checkmatrix(mat1);
   p=get_matrix(mat1,&r1,&c1);
  gyo=*(stack-1);

  k=SINT_NUMB(gyo);
  r2                 = k;
  r3=r1-r2;
  if((k<1)||(r3<1))  error(E_range, gyo);
  mat2 = makematrix(r2,c1, NIL);
  psh(mat2);
  mat3 = makematrix(r3,c1, NIL);
   p=get_matrix(mat1,&r1,&c1);
   q=get_matrix(mat2,&r2,&c2);
   r=get_matrix(mat3,&r3,&c3);
  psh(mat3);
  z=CONS(mat3,a_nil);
   psh(z);
  for(i=0;i<k;i++){
    for(j=0;j<c1;j++){
      *q++=*p++;
    }
  }
  for(i=0;i<r3;i++){
    for(j=0;j<c1;j++){
      *r++=*p++;
    }
  }
  z=CONS(mat2,z);
  stack-=3;
  return z;
}

ptr s_matrix_divide_tr(int narg, ptr env)
{
  ptr  mat1,mat2,mat3,gyo,z;
  int i,j,k,r1,c1,r2,c2,r3,c3;
   ptr *p,*q,*r;
    checknarg(narg-2);
  mat1=*(stack-2);
  checkmatrix(mat1);
   p=get_matrix(mat1,&r1,&c1);
  gyo=*(stack-1);

  k=SINT_NUMB(gyo);
  c2                 = k;
  c3=c1-c2;
  if((k<1)||(c3<1))  error(E_range, gyo);
  mat2 = makematrix(r1,c2, NIL);
  psh(mat2);
  mat3 = makematrix(r1,c3, NIL);
   p=get_matrix(mat1,&r1,&c1);
   q=get_matrix(mat2,&r2,&c2);
   r=get_matrix(mat3,&r3,&c3);

  psh(mat3);
  z=CONS(mat3,NIL);
   psh(z);
  for(i=0;i<r1;i++){
    for(j=0;j<c2;j++){
      *q++=*p++;
    }
     for(j=0;j<c3;j++){
      *r++=*p++;
    }
  }
  z=CONS(mat2,z);
  stack-=3;
  return z;
}

ptr s_matrix_rotate(int narg,ptr env)
{
  ptr x;
  int n,m,sh,start,end;
  ptr *p,*work;
   if(narg==2){
      x=*(stack-2);
      sh=SINT_NUMB(*(stack-1));
      start=0;
   }
  else if(narg==3) {
      x=*(stack-3);
      sh=SINT_NUMB(*(stack-2));
      start=SINT_NUMB(*(stack-1));
   }
  else{
    checknarg(narg-4);
      x=*(stack-4);
      sh=SINT_NUMB(*(stack-3));
      start=SINT_NUMB(*(stack-2));
      end=SINT_NUMB(*(stack-1));
  }
   checkmatrix(x);
    p=get_matrix(x,&n,&m);
    if((start<0)||(start>=n)) error(E_range,MAKESINT(start));
    if(narg<4)
          end=n;
    else
       if((end<=start)||(end>n)) error(E_range,MAKESINT(end));
    sh%=n;
    work=getwork(m);
    p+=start*m;
    n=end-start;
    if(sh<0) goto left;
    while(sh--)
      matrix_rotate_right(p,n,m,work);
    return x;
left:
     sh=-sh;
       while(sh--)
      matrix_rotate_left(p,n,m,work);
    return x;
}

void matrix_rotate_left(ptr *p,int n,int m,ptr * work)
{
   int size;
   size=m*sizeof(ptr);
   memcpy(work,p,size); /***  0 gyo->work ***/
   while(--n){
    memcpy(p,p+m,size);
    p+=m;
   }
  memcpy(p,work,size);
 return;
}

void matrix_rotate_right(ptr *p,int n,int m,ptr * work)
{
   int size;
   size=m*sizeof(ptr);
   p+=(n-1)*m;
   memcpy(work,p,size); /***  last gyo->work ***/
   while(--n){
    memcpy(p,p-m,size);
    p-=m;
   }
  memcpy(p,work,size);
 return;
}

ptr s_matrix_vector_set(int narg,ptr env)
{
  ptr  mat,  vect;
  ptr s,m,z,mat3;
  int gyo,i,j,k,r,c;
   ptr *p,*q;
  checknarg(narg-3);
   mat=*(stack-3);
   gyo=SINT_NUMB(*(stack-2));
   vect=*(stack-1);
   checkmatrix(mat);
   checkvect(vect);

   p=get_matrix(mat,&r,&c);
   q=get_vector(vect,&k);
  if((gyo<0)||(gyo>=r))  error(E_range, *(stack-2));
  p+=gyo*c;
  if(k<c)  c=k;
  while(c--)
     *q++=*p++;
  return vect;
}

ptr s_matrix_vector_set_m(int narg,ptr env)
{
  ptr  mat,  vect;
  ptr s,m,z,mat3;
  int gyo,i,j,k,r,c;
   ptr *p,*q;
  checknarg(narg-3);
   mat=*(stack-3);
   gyo=SINT_NUMB(*(stack-2));
   vect=*(stack-1);
   checkmatrix(mat);
   checkvect(vect);

   p=get_matrix(mat,&r,&c);
   q=get_vector(vect,&k);
  if((gyo<0)||(gyo>=r))  error(E_range, *(stack-2));
  p+=gyo*c;
  if(k<c)  c=k;
  while(c--)
     *p++=*q++;
  return mat;
}



