
;;;;  compile-file
;;;;
;;;(in-package "compiler")
(load "opt")
(load "imm-final")
(load "compiler-load")
(load "vm-op-table")

(define *compile-name* "")  ;;; reverse.scm-> "reverse"
;;;(define *comp-out-file* "")

(define (make-name-sr2 file)
   (let* ((s (file-sepa-ext file)))
     ;;(tk-string (car s) ".sr2")
     (string-append (car s) ".sr2")
  )
)

(define *comp-buff* (make-string 1000))

(define (load-form s)
  (let ((out  (open-output-string *comp-buff* ))(in 0) (e 0) )
    (write s out)
    (close-port out)
    (set! in (open-input-string *comp-buff*))
    ;;;(catch (load in))   ;;;;;; 04.5.17
    ;;(load in)   ;;;;;;toriaezu
    (set! e (catch (load in)))   ;;;;;; 04.5.17
    (when e  ;;; error!
       (format #t "~%~%compiler load fail: ~s~%" s)
       (format #t "compile-file load-form load fail ~%" )
       (bye)
       (error 8888 s)
    )
    (close-port in)
 )    
)

(define (special-begin? x)
     (set! x (special? x))
    (if (eq? x 'begin)
        x 
        #f
    )
)

(define (makesr2  file)
  (let* ((s 0)(f 0)(out #f))
     (set! s (open-input-file  file ))
     (set! out (open-output-file "aout.sr2"))
         (format  #t "out ~s~%" out)
        (write 'BEGIN-SR2 out)
        (newline out)
         (do ((x (read s)(read s)))
           ((eof-object? x)
                  (close-input-port  s)
                )
             (set! x (macro-expand-rec x ()))
         ;;;(format  #t "expan x ~s~%" x)
             (compiler-eval-package  x)
            (format #t "~s~%" x)
            (set! x (compile x ()))   ;;;
            (format #t "compiled:~s~%" x)
            (format #t "opt:~s~%" x)
            (set! x (syntax->data x))   ;;;V2.0  to be more exact!
            (set! x (c-opt x))    ;;;;  optimize
            (set! x (imm-final x))   ;;;;  IMM-t  etc.
             (format out "~s~%" x)
           )
        (close-output-port  out)
))

(define (comp-load file) 
   (let* ((s 0)(f 0))
     (set! s (open-input-file  file ))
         (do ((x (read s)(read s)))
           ((eof-object? x)
                  (close-input-port  s)
                )
           (when  (pair? x)
              (set! f (special? (car x)))
              (if (special-begin? (car x))
                 (compiler-eval-define-begin  x );; pass org
                 (compiler-eval-define x (macro-expand-rec x ()))
                                     ;; pass org&expand
               )
            )
           )
))

(define (compiler-eval-define-begin x)
   (let* ((f 0))
    (set! x (cdr x))
    (while (pair? x)
        (set! f (car x))
        (if (pair? f)
          (if (special-begin? (car f))
            (compiler-eval-define-begin f) 
            ;;;(compiler-eval-define f  (macro-expand-rec x ())) ;;; toriaezu
            (compiler-eval-define f  (macro-expand-rec f ())) ;;; toriaezu
          )
        )
        (set! x (cdr x))
    )
))

(define (compiler-eval-define org x)   ;;; x is expanded of org
  (if (pair? x)
     (let* ( (f (car x))(sp 0)(g (car org)))
         ;;;  f may be alias
         (set! sp (special? f))
         (if sp 
             (set! f sp)
         )
      #|
        (when (eq? 'define-syntax (car org))
              (set! f 'define-syntax)
                   ;;;; define-syntax is expanded (define ...)
                              ;;; so we need car  org
         )
      |#
       (format #t "compiler-eval-define ~s ~s~%" f x) ;;;; for debug

        (cond ((eq? f 'begin)
                  (compiler-eval-define-begin x)
                )
              ((eq? f 'load)
                          (compiler-load-define (cadr x)))  ;;;;load file 04.5.3
                         ;;; (comp-load (cadr x) #f)
              ((eq? g 'defstruct) (load-form  org))  ;;;;needs! 04.5.3
              ((eq? g 'compiler-eval) ;;;other of defstruct
                    (load-form (cons 'begin (cdr org)))
                 )
              ((eq? f 'compiler-eval) ;;;other of defstruct
                    (load-form (cons 'begin (syntax->data (cdr x))))
                    ;(load-form (cons 'begin (cdr org)))
               )
              ((eq? g 'compiler-not-eval) )
             ;; ((eq? f 'define) (define-load-form  x))
              ((eq? g 'define-macro) (load-form  org))
              ((eq? g 'define-syntax) (load-form  org))  ;;;;needs! 04.5.3
             ;; ((eq? f 'defstruct) (load-form  org))  ;;;;needs! 04.5.3
              ((eq? f 'require) (load-form  org))
              ((eq? f 'provide) (load-form  org))
              ((eq? f 'remove-module) (load-form  org))
              (else
                 (compiler-eval-package org ) )
          )
      )
  )
)

(define (compiler-eval-package x)
  (if (pair? x)
     (let* ((f (car x)))
        (cond ((eq? f 'begin)
                    (set! x (cdr x))
                 (while (pair? x)
                    (compiler-eval-package (car x))
                    (set! x (cdr x))
                 )
                )
                         ;;; (comp-load (cadr x) #f)
              ((or  (eq? f 'in-package) (eq? f 'export)(eq? f 'use-package)
                    (eq? f 'intern)(eq? f 'shadowing-intern)(eq? f 'unintern)
                    (eq? f 'unuse-package))
                    ;;;;(catch (eval x))
                     (eval x)  ;;; toriaezu
                  )
              (else )
          )
      )
  )
)
                     

(define (find-scm file)
  (let ((s 0))
  (set! file (find-source file #t))   ;;; s_load.c
  (if file (set! *compile-name* (car (file-sepa-ext file)))
        (error 1222 file)  )
   file
))


(define (compiler-load files)
  (while (pair? files)
    (comp-load (pop files) )
))

(define (compile-file file)
     ;;;(set! *verbose* #f)
     (comp-load file )
    ;;; (init-compile)
     ;;;(in-package "compiler")
    ;;; (set! *verbose* #t)
     (makesr2 file )
     (rename-file "aout.sr2" (make-name-sr2  file ))
     '***end-of-compile-file***
)
;;;  2010.9.1
;;;   syntax-symbol -> org symbol

(define (syntax->data  x)
   (if (syntax-symbol? x)
       (car (syntax-cdr x))
      (if (pair? x)
         (cons (syntax->data (car x)) (syntax->data (cdr x)))
         x
      )
   )
)
           
                 
