;;; 変数の最適化
;;;

;;;; 変数のアクセス記録を作る
;;;; 変数の種類
;;;  :TOP  :ENV
;;;  さらに、　:ENVには
;;;   :BIND :LAMBDA
;;;   がある。
;;;  さらに、:BIND にはどのBlockの変数か？
;;;  そして、アクセスは、どのBLOCKからかがある。
;;;
;;; そこで、次のようにする。
;;;
;;; ( var   define-flag  load-access-count  set-access-count)
;;;   define-flag は:warningで判定する。
;;; (( var .  block) :BIND   load-access set-access)
;;; (( var .  block) :LAMBDA  load-access set-access)
;;; ただし、:LETBIND , :BIND はset-accessにはcountしない。
;;;;
;;; このdata を、*var-access*に作る。
;;; また、:LOAD,:SET の:ENV 変数を(x . block)に変換する。
;;; :ENV 変数の値のdatabase も作る。
;;;  ()  unknown
;;;  (:IMMIDIATE  *)
;;;; (:KNOWN #t)

(in-package "compiler")

#|
(define (dummy? scm)
  (and (pair? scm) (eq? (cadr scm) *dummy-func*))
)
|#
(define (make-var-access access)
   (let ((v *comp-vector*)(n *comp-size*)(x 0)
            (scm ()) (top 0)(block 0)(var 0) (env ())(env-stack ()))
      (set! *var-value* ())
      (set! *letrec-var* ())
     (do ((i 0))
         ((<= n i) )
         (set! x (vector-ref v i))
         (set! top (car x))
         (cond ((eq? top :BIND)
                 (set! var (cadr x))
                 (when (symbol? var)
                     (set! var (cons var block))
                     (vector-set! v i (list :BIND var))
                 )
                ;; (if (dummy? scm)
                ;;     (begin
                 ;;     (set! access (cons (list   var  :BIND 0 -1) access));;;letrec bind
                 ;;     (set! *letrec-var* (cons var *letrec-var*))
                 ;;    )
            (set! access (cons (list   var  :BIND 0 0) access))
                ;;; )
                 (set-var-value  var scm)
                 (set! env (cons  var env))
                 )
               ((eq? top :LOAD)
                  (set! access (load-access x v i  access env))
                  (set! var (cdr access))
                  (set! access (car access))
                  (set! scm (load-var-value x var))
                 )
               ((eq? top :SET)
                  (set! access (set-access x v i access env))
                  (set! var (cdr access))
                  (set! access (car access))
                  (set-var-value  var scm)
                 )
               ((eq? top :DEFINE)
                   (set! var (cadr x))
                   (set! access (cons (list var #t  0 0) access))
                 )
               ((eq? top :LETBIND)
                  (do ((vars (cadr x)))
                      ((null? vars))
                        (set! var (cons (car vars) block))
                        (set! access (cons (list  var :BIND 0 0) access))
                         (set! env (cons  var env))
                        (set! vars(cdr vars))
                    )
                   (set! scm ())
                 )
               ((or (eq? top :BLOCK-IN) (eq? top :LAMBDA-BEGIN))
                  (set! env-stack (cons (cons  env block) env-stack))
                  (set! block (cadr x))
                )
               ((or (eq? top :BLOCK-OUT)(eq? top :LAMBDA-END))
                  (set! var (car env-stack))
                  (set! env (car var))
                  (set! block (cdr var))
                  (set! env-stack (cdr env-stack))
                  (if (eq? top :LAMBDA-END) (set! scm (list :LAMBDA)))
                )
                ((eq? top :LAMBDA-BIND)
                  (do ((vars (cadr x)))
                      ( (null? vars) )
                         (if (pair? vars)
                             (begin
                                (set! var (car vars))
                                (set! vars(cdr vars))
                             )
                             (begin
                                (set! var vars)
                                (set! vars ())
                             )
                          )
                       (when var   ;;; may be #f,  by frame-var
                          (set! access (cons (list (cons var  block) :LAMBDA 0 0) access))
                          (set! env (cons (cons var block) env))
                       )
                    )
                   (set! scm ())
                 )
               ((eq? top :IF)
                  (if (eq? (cadr x) :T)
                        (set! scm (list :IMMIDIATE #f))
                        (set! scm (list :KNOWN #t))
                  )
                 )
               ((or (eq? top :STACK-NARG) (eq? top :FUNC)(eq? top :FUNC-TAIL) )
                   )
               (else (set! scm ()) )
          )
         (set!  i(+ i 1))
     )
   (set! *var-access* access)
   *var-access*
 )
)
(define (set-var-value var scm)
  (let ((s 0))
    (set! s (assoc var *var-value*))
    (if s
         (set-cdr! s scm)
         (set! *var-value*  (cons (cons var scm) *var-value*))
    )
))

(define (load-var-value x var)
  (let ((s 0))
    (if  var
       (begin
          (set! s (assoc var *var-value*))
          (if s (cdr s)
                ()
          )
        )
       (cdr x)
    )
))

(define (load-access x v ind access env)
  (let ((ac (cadr x)) (var #f) (c 0)(data 0))
    (when (eq?  ac :TOP)
       (set! var (caddr x))
       (set! data (assq  var access))
       (if data
            (load-inc data)
            (set! access (cons (list var #f  1  0) access)) ;;;; TOP
       )
    )
    (when (eq?  ac :ENV)
       (set! var (caddr x))
       (when (symbol? var)  ;;; Not changed
           (set! data (assq var env))
           (if data
               (vector-set! v ind (list :LOAD :ENV data)) ;;; change (x . block)
               (error 999  x)   ;;;Never occur
           )
          (set! var data)
       )
       (set! data (assoc  var access))
       (if data
              (load-inc data)
              (error 999  x)   ;;;Never occur
        )
    )
   (cons access var)
))

(define (set-access x v ind access env)
  (let ((ac (cadr x)) (var :FRAME) (c 0)(data 0))
    (when (eq?  ac :TOP)
       (set! var (caddr x))
       (set! data (assq  var access))
       (if data
            (set-inc data)
            (set! access (cons (list var #f  0  1) access))
       )
    )
    (when (eq?  ac :ENV)
       (set! var (caddr x))
       (when (symbol? var)  ;;; Not changed
           (set! data (assq var env))
           (if data
               (vector-set! v ind (list :SET :ENV data)) ;;; change (x . block)
               (error 999  x)   ;;;Never occur
           )
          (set! var data)
       )
       (set! data (assoc  var access))
       (if data
              (set-inc data)
              (error 999  x)   ;;;Never occur
        )
    )
   (cons access var)
))

(define (load-inc data)
   (let ((c 0))
      (set! data (cddr data))
      (set! c (car data))
      (set-car! data (+ c 1))
))

(define (set-inc data)
   (let ((c 0))
      (set! data (cdddr data))
      (set! c (car data))
      (set-car! data (+ c 1))
))

