;;;   04.4.23
;;;
;;;;  for GCL
(define pi 3.14159265358979)
(define nil ())

(define-syntax progn
   (syntax-rules ()
     ((progn body ...) (begin body ...))
   )
)

(define-syntax defvar
  (syntax-rules ()
    ((defvar x) (define x #f))
    ((defvar x v) (define x v))
  )
)

(define-syntax defconstant
   (syntax-rules ()
     ((defconstant x v)
      (begin
        (if (symbol-bind? 'x)
            (format #t "warning: constant variable redefined!!   ~s~%" 'x)
         )
        (define x v) (constant! 'x))  )
 ) 
)


(define-syntax setq
   (syntax-rules ()
     ((setq a b) (set! a b))
     ((setq a b c ...)
       (begin (set! a b) (setq c ...)))
   )
)

(define-syntax terpri
   (syntax-rules ()
     ((terpri) (newline))
   )
)

(define-syntax null
   (syntax-rules ()
     ((null x) (null? x))
   )
)

(define-syntax eq
   (syntax-rules ()
     ((eq x y) (eq? x y))
   )
)

(define-syntax equal
   (syntax-rules ()
     ((equal x y) (equal? x y))
   )
)

(define-syntax nreverse
   (syntax-rules ()
     ((nreverse y) (reverse y))
   )
)

(define-syntax incf
   (syntax-rules ()
     ((incf x) (set! x (+ 1 x)))
     ((incf x v)(set! x (+ x v)))
   )
)

(define-syntax decf
   (syntax-rules ()
     ((decf x) (set! x (- x 1 )))
     ((decf x v)(set! x (- x v)))
   )
)

#|
(define (intern x)
  (string->symbol x))
|#


(define-syntax defun     ;;;;  enable return-from function
   (syntax-rules (return-from let-syntax syntax-rules)
     ((defun func para body ...)
        (define func (lambda para
              (call/cc (lambda (return)
                 (let-syntax 
                   ((return-from (syntax-rules ()
                       ((return-from  f value)
                          (if (eq? 'f 'func) (return value) (error 9999 'f)) )
                       )
                     )
                   )
                  body ...
                  )
                ) )
            )
         )   )
   )
)
#|
(defun f(x)
  (if (= x 1) (return-from f 'first))
  (if (= x 2) (return-from f 'second))
  (if (= x 3) (return-from f  'error))
  888
)

(let* ((x 7)) 888)

|#


#|
(define-syntax loop (syntax-rules (return)
  ((loop s ...)
    (call/cc (lambda (return)
        (letrec ((repeat (lambda () s ... (repeat))))
            (repeat)
        )
      )
     )
   )
 )
)
|#

(define-syntax loop (syntax-rules (let-syntax syntax-rules return )
  ((loop s ...)
    (call/cc (lambda (out)
        (let-syntax 
          ((return (syntax-rules ()
             ((return) (out #t))
             ((return val)(out val))
            )  ) )
        
          (letrec ((repeat (lambda () s ... (repeat))))
            (repeat)
           )
        )
      )
     )
   )
 )
)



(define-syntax multiple-value-setq (syntax-rules ()
  ((multiple-value-setq (a) form)
      (let* ((temp form))
         (set! a temp) temp ))
  ((multiple-value-setq (a vars ...) form)
     (let* ((temp (values->list form)))
       (if (null? temp)
           (begin
              (set! a #f)
              (multiple-value-setq "false" vars ...) 
              #f
           )
           (begin
              (set! a (car temp))
              (set! temp (cdr temp))
              (multiple-value-setq  temp vars ...)
              a
           )
        )
       ))
   ((multiple-value-setq "false")
        #f
     )
   ((multiple-value-setq "false" a vars ...)
      (begin
         (set! a #f)
         (multiple-value-setq "false" vars ...)
      )  )
   ((multiple-value-setq value)
        #f)
   ((multiple-value-setq value a vars ...)
          (if (null? value)
                 (multiple-value-setq  "false" a vars ...)
               (begin
                 (set! a (car value))
                 (set! value (cdr value))
                 (multiple-value-setq value  vars ...)
               )     
           )
        )
   )
)

(define-syntax aref
  (syntax-rules ()
    ((aref vec n)(vector-ref vec n))   
  )
)

(define (list-set list pos val)
   (let ((s()))
     (loop
        (if (= pos 0)(return))
        (push (pop list) s)
        (decf pos)
     )
    (pop list)
     (push val list)
    (loop
       (if (null? s) (return))
       (push (pop s) list)
     )
    list
))

(define-syntax setf
  (syntax-rules (aref nth)
    ((setf (aref vec n) val)(vector-set! vec n val))
    ((setf (nth lis n) val)(list-set lis n val))
    ((setf var val)   (set! var val))
  )
)

(define-syntax swap
  (syntax-rules ()
    ((swap x y)(let* ((temp x)) (set! x y)(set! y temp))) 
  )
)

(define-syntax dotimes
  (syntax-rules (return)
    ((dotimes (var count) body ...)
       (let* ((c count)(var 0))
         (loop
            (if (<= c var) (return ()))
             body ...
            ;;;(set! c (- c 1))
            (set! var (+ var 1))
          )
       )
      )
    ((dotimes (var count form) body ...)
       (let* ((c count)(var 0))
         (loop
            (if (<= c var) (return form))
             body ...
            (set! var (+ var 1))
          )
       )
      )
  )
)

(define-syntax dolist
  (syntax-rules (return)
    ((dolist (var lis) body ...)
       (let* ((c lis)(var c))
         (loop
            (if (null? c) (return ()))
            (set! var (car c))
             body ...
            ;;;(set! c (- c 1))
            (set! c (cdr c ))
          )
       )
      )
    ((dolist (var lis form) body ...)
       (let* ((c lis)(var c))
         (loop
            (set! var ())
            (if (null? c) (return form))
            (set! var (car c))
             body ...
            (set! c (cdr c ))
          )
       )
      )
  )
)



;;; property list
;;;;
(define *prop-list* ())

(define (set-prop x prop-name prop)
   (if (get-prop x prop-name)
     (remove-prop x prop-name))
   (push (cons x (cons prop-name prop)) *prop-list*)
)

(define (get-prop x prop-name)
   (let ((s 0)(res #f))
     (do ((plist *prop-list*))
         ((null? plist))
        (set! s (pop plist))
        (and (eq? (car s) x) (eq? (cadr s) prop-name)
           (begin
              (set! res (cddr s))
              (set! plist ())
            )
        )
      )
    res
))

(define (remove-prop x prop-name)
  (let ((s 0)(res ()))
     (do ((plist *prop-list*))
         ((null? plist))
        (set! s (pop plist))
        (if (not (and  (eq? (car s) x) (eq? (cadr s) prop-name)))
           (push s res)
        )
      )
    (set! *prop-list* res )
))

;;;;

#|
(let ((s 1)(i 0))
   (loop (if (= i 100) (return s)) (set! i (+ i 1)) (set! s (* s i)))
)
(s '(multiple-value-setq (x y z) (values 2 3 4)))

(define (fact n)
 (let ((s 1))
  (loop
    (if (= n  1) (return s))
   (set! s (* s n))
    (set! n (- n 1 ))
   ;(format #t "~s~%" s)
  )
))

|#

