;;;  Regexp  Reader
;;;  2005.8.23
;;;  Ver2.3
;;;    Stack  version

(in-package "regexp")

(require "allow-unread")

(define *auto-regexp* #f)
(define *free-space* #f)

(define *if-num* -10000)  ;;; use only here
(define *nocap-num* 0)  ;;; use only here
;;;

(define-syntax push-#f
   (syntax-rules ()
      ((push-#f res)
             (if (eq? (car res) #t)
                    (set! res (cdr res))
                    (set! res (cons #f res))
             )
       )
   )
)
                                                                                          
;;;;(define in (open-input-func input-string "step"))
(define R-getc  read-char )
(define  R-ungetc  unread-char)
(define R-peekc  peek-char)
;;;;;

(define R-table (make-vector 256 (cons 0 0)))

(define RS-table R-table)
(define RL-table (make-vector 256 (cons 0 0)))
(define RU-table R-table)

(define (set-R-table c attr func)
   (vector-set! R-table (char->integer c) (cons attr func))
)


(define (make-regexp pat)
  (let* ((s 0)(c 0))
      ;;;;;(set! pat (open-input-str  pat))
      ;;;(set!  pat  (open-input-func input-string pat))
      (set!  pat  (allow-unread-str-port pat))
      (set! s (R-read pat ) )
      (close-port pat)
      (set! c (R-count s ) )
      (if c 
         (set! c (cons c (R-check-end s)))
         (set! c (cons c #f))
     )
      (cons (list *group-num* *named-group* (R-mincount  s) c (R-check-str s))  s )
))
                                                                                            
(define (r-esc pat)
  (let* ((x (R-getc pat)))
   (cond ((eq? x #\n) #\newline)
         ((eq? x #\t) #\tab)
         ((eq? x #\e) #\escape)
         ((eq? x #\f) #\page)
         ((eq? x #\r) #\return)
         ((eq? x #\a) #\bell)
         ((eq? x #\v) #\vt)
         ((eq? x #\c) #\kspace)
         ((eq? x #\s) 'r-space?)
         ((eq? x #\S) 'not-space?)
         ((eq? x #\d) 'r-digit?)
         ((eq? x #\D) 'not-digit?)
         ((eq? x #\w) 'r-word?)
         ((eq? x #\W) 'not-word?)
         ((eq? x #\k) 'r-kanji?)
         ((eq? x #\K) 'not-kanji?)
         ((eq? x #\/) #\\)
         ;;;  new Ver2.2
         ((eq? x #\<) (list 'r-pos :ws))
         ((eq? x #\>) (list 'r-pos :we))
         ((eq? x #\b) (list 'r-pos :wse))
         ((eq? x #\B) (list 'r-pos :wsenot))
         ((eq? x #\G) (list 'G-pos ))
         ((eq? x #\g) (list 'g-pos ))
         ((eq? x #\!) *fail*)
         ((eq? x #\#) #f)
         ((eq? x #\A) (list 'r-pos 'abs-top))
         ((eq? x #\z) (list 'r-pos 'abs-end))
         ((eq? x #\c) (set! x (R-getc pat))
                 (integer->char (- (char->integer x) (char->integer #\A) -1))
                )
         ((eq? x #\Q) 
                (set! RU-table R-table)
                (set! R-table  RL-table)
                #f)
         ((eq? x #\x) 
             (integer->char (r-hexnumber pat))
           )
         ((char-numeric? x)
                 (R-ungetc x pat)
                 (list 'group-ref (r-number pat))
           )
         (else x)
   )
))

(define (r-esc-literal  pat)
  (let* ((x (R-peekc pat)))
     (if (eq? x #\E)
         (begin
             (R-getc pat)
             (set! R-table  RU-table)
              #f
         )
         #\\
     )
))

(define (r-range? x)
    (or (eq? x :range) (eq? x :range+)(eq? x :range?))
)

(define (r-range pat)
   (let* ((x 0) (n (r-number pat))(m n))
      (set! x (R-getc pat))
      (when (eq? x #\,)
          (set! m (r-number pat))
          (set! x (R-getc pat))
      )
     (if (not (eq? x #\})) (error 467 x))
      (set! x (R-peekc pat))
    (if (eq? x #\?)  ;;;;  hi-yokubari
      (begin
         (R-getc pat)
         (list :range? (cons n m))  ;;;;  
      )
     (if (eq? x #\+)  ;;; atomic
       (begin
         (R-getc pat)
         (list :range+ (cons n m))  ;;;;
      )
      (list :range (cons n m))  ;;;;  
    )
  )
))
         

(define (r-number pat)
   (let*  ((n 0)(x (R-getc pat)) (zero (char->integer #\0)))
     (while (char-numeric? x)
        (set! n (+ (* n 10)   (- (char->integer x)  zero)))
        (set! x (R-getc pat) )
     )
   (R-ungetc x pat)
   n
))

(define (char-hex->num x)
   (let*  ((n 0) (zero (char->integer #\0))(ten  (char->integer #\A)))
     (set! n  (char->integer x))
     (cond ((and (>= n zero) (< n (+ zero 10)))
               (- n zero) )
           ((and (>= n ten) (< n (+ ten 6)))
               (- n ten -10)  )
           (else
              #f
            )
     )
))
               
(define (r-hexnumber pat)
   (let*  ((n 0)(k 0)(x (R-getc pat)) (zero (char->integer #\0)))
     (while (begin (set! k (char-hex->num x)) k)
        (set! n (+ (* n 16)   k))
        (set! x (R-getc pat) )
     )
   (R-ungetc x pat)
   n
))


(define (r-esc-bag  pat)   ;;;;   in [....]
  (let* ((x (R-getc pat)))
   (cond ((eq? x #\n) #\newline)
         ((eq? x #\t) #\tab)
         ((eq? x #\e) #\escape)
         ((eq? x #\f) #\page)
         ((eq? x #\r) #\return)
         ((eq? x #\a) #\bell)
         ((eq? x #\v) #\vt)
         ((eq? x #\c) #\kspace)
         ((eq? x #\s) 'r-space?)
         ((eq? x #\S) 'not-space?)
         ((eq? x #\d) 'r-digit?)
         ((eq? x #\D) 'not-digit?)
         ((eq? x #\w) 'r-word?)
         ((eq? x #\W) 'not-word?)
         ((eq? x #\k) 'r-kanji?)
         ((eq? x #\K) 'not-kanji?)
         ((eq? x #\/) #\\)
         ;;;  new Ver2.2
        ;;; ((eq? x #\<) (list 'r-pos :ws))
        ;;; ((eq? x #\>) (list 'r-pos :we))
         ((eq? x #\b) #\backspace)
        ;;;; ((eq? x #\B) (list 'r-pos :wsenot))
        ;;; ((char-numeric? x)
        ;;;         (list 'group-ref (- (char->integer x)  (char->integer #\0)))
        ;;;   )
         (else x)
   )
))


   
(define (R-char? x)
   (cond ( (keyword? x) #f)
         ((pair? x) 
             (eq?  (car x) 'r-bag) 
              ;;;;(eq?  x 'r-pos) (eq? x 'group-ref))
           )
        ( (boolean? x) #f)
        (else #t)
   )
)

(define (R-atom? x)
  (and  (pair? x) (symbol? (car x)))
)

(define (R-pop res)
  (let* ((x (car res)))
    (cond ((char? x) (R-pop-pure res))
          ((R-char? x) (R-pop-genstr res))
          (else res)
    )
))

(define (R-pop-genstr res)
  (let* ((x (pop res))(str ())(cont #t))
          (push x str)
          (while (and (pair? res) cont)
             (set! x (car res))
             (if (and (R-char? x) (not (char? x)))
                 (begin
                    (push x str)
                    (set! res (cdr res))
                 )
                 (set! cont #f)
             )
          )
         (cons (cons 'string  str) res)
))

(define (R-pop-pure res)
  (let* ((x (pop res))(str ())(cont #t))
          (push x str)
          (while (and (pair? res) cont)
             (set! x (car res))
             (if (char? x)
                 (begin
                    (push x str)
                    (set! res (cdr res))
                 )
                 (set! cont #f)
             )
          )
         (cons (list 'pure-string  (list->string str))  res)
))

(define (R-left? x)
  (and (pair? x)(eq? (car x) 'g-top))
)

(define (R-right? x)
  (and (pair? x)(eq? (car x) 'g-end))
)

(define (R-pop-group res)    ;;;  left  must be R-left  (g-top n)
  (let* ((left (pop res))(ans (list left))(x 0)
             (n (cadr left))(right (list 'g-end n)) (cont #t) (inside (cadr left)) )
        (set! inside (and (number? inside)(< inside 0)))   ;;; (g-top -n)
         (if inside (set! ans ()))
       (while (and (pair? res) cont)
            (set! x (pop res))
            (push x ans)
           ;;;(set! ans (append x ans))
            (if (equal? x right)
                 (set! cont #f)
           )
       )
      (if inside (pop ans))
  (cons (reverse ans)  res)
))
;;;;   for if and other?  left,right are omitted!
(define (R-pop-group-inside  res)    ;;;  left  must be R-left  (g-top n)
  (let* ((left (pop res))(ans ())(x 0)
             (n (cadr left))(right (list 'g-end n)) (cont #t) )
       (while (and (pair? res) cont)
            (set! x (pop res))
           ;;;(set! ans (append x ans))
            (if (equal? x right)
                 (set! cont #f)
                 (push x ans)
           )
       )
  (cons (reverse ans)  res)
))

;;;;  these information are used in regexp 
(define *group-num* 0)
(define *named-group* ())
              
(define (R-read  pat )
  (let* ((res ())(x 0) (cont #t))
     (set! *group-num* 0)
     (set! *nocap-num* 0)
     (set! *if-num* -10000)
     (set! *named-group* ())
     (set! R-table (vector-copy RS-table))
     (R-after (R-read-stack (R-read-pre pat)))
))

(define (R-after ans)   ;;;; take :begin
   (if (pair? ans)
        (let* ((x (pop ans)))
           (if  (pair? x) 
              (if (eq? (car x) :begin) 
                     (append (R-after (cdr x)) (R-after   ans))
                   (if (symbol? (car x))
                     ;;;(cons   x (R-after ans))
                     (cons   (group-ref-cut x) (R-after ans))
                     (cons  (R-after x) (R-after ans))
                   )
             )
            (cons x (R-after ans))
           )
       )
     ans
  )
)
;;;   find short-cut from name of group 
(define (group-ref-cut x)
    (if (and (eq? (car x) 'group-ref) (symbol? (cadr x)))
      (let* ((name (cadr x)))
          (set! name (assq name *named-group*))
           ;;;; (format #t "name ~s~%" name)
          (if (and name (null? (cddr name)))
              (list 'group-ref (cadr name))
              x
          )
      )
      x
   )
)

(define (R-disp func pat)   ;;;  hi-yokubari
  (let* ((x (R-peekc pat)) (new 0))
    (set! new
      (cond ( (eq? x #\?)
                (cond ((eq? func :*) :*?)
                      ((eq? func :+) :+?)
                      ((eq? func :?) :??)
                      (else func)
                )
               )
            ( (eq? x #\+)
                (cond ((eq? func :*) :*+)
                      ((eq? func :+) :++)
                      ((eq? func :?) :?+)
                      (else func)
                )
               )
            (else func)
      )
    )
   (if (eq? func new) 
           func
        (begin
           (R-getc pat)
           new
       )
   )
))
       
        
;;;;  #t ...  #f are marker for (,|,)              
(define (R-read-pre  pat )
  (let* ((x 0)(res ()) (cont #t) (k 0)(gr *group-num*) (code 0)(attr 0)(func 0)
         (kstack ()) (gnum 0)(test 0)(back-f #f))
        (set! x (R-getc pat))
        (push #t res)
    (while (not (eof-object? x))
        ;;;(format #t "sub x ~s~%" x)
        (set! code (char->integer x))
        (if (< code 256)
           (begin
             (set! k (vector-ref R-table code)) 
             (set! attr (car k))
             (set! func  (cdr k))
           )
           (set! attr 0)
       )
        (cond ((eq? attr 0) (push x res))
              ((eq? attr 8)  (push func res))  ;;; r-dot  . 
            ;;;;   . *,+,?  ,  {n,m}
              ((or (eq? attr 1) (eq? attr 9))
                      (if (eq? attr 1)
                          (set! func (R-disp func  pat))
                          (set! func (funcall func  pat))
                      )
                      (if (eq? (car res) #t)    ;;; case of (...)*abc
                         (begin
                            (pop res)
                             (push func res)
                             (push #t res)
                         )
                         (push func res)
                      )
                  )
            ;;;   |
              ((eq? attr 2) 
                       (push-#f res)
                       (push func res)
                       (push #t res)
                    )
            ;;;  (    
              ((eq? attr 3 )
                  (set! gnum (group-pre pat))
                   (if (number? gnum )
                      (begin
                         (push-#f res)
                         (push  gnum kstack)
                         (push (list func gnum) res)
                         (push #t res)
                      )
                     (when gnum
                        ;;;;(format #t "gnum ~s~%" gnum)
                        (if (pair? gnum)
                          (begin 
                              (set! k (car gnum ))
                              (cond  ((eq? k :if)
                                        (set! test (caddr gnum))
                                        (set! gnum (list (cadr gnum)))  
                                        (push-#f res)
                                        (push :if res)
                                        (push #t res)
                                        (push (list 'if-test test) res)
                                         (push #f res)
                                        (if (eq? :test test)
                                              (push :test  res)
                                         )
                                        (push  gnum kstack)
                                        (push (list func gnum) res)
                                         (push #t res)
                                      )
                                     ((or (eq? k :forward) (eq? k :backward))
                                                  ;;;  as :not
                                          (when (keyword? (cadr gnum))  ;;; :not
                                              (set! gnum (list k (caddr gnum)))
                                              (push-#f res)
                                              (push :not res)
                                              (push #t res)
                                          )
                                         (push-#f res)
                                         (push k res)
                                         (set! gnum (cdr gnum))
                                         (push  gnum kstack)
                                         (push (list func  gnum) res)
                                         (push #t res)
                                     )
                                     ((eq? k :atomic)   ;;;  as :not
                                         (push-#f res)
                                         (push k res)
                                         (set! gnum (cdr gnum))
                                         (push  gnum kstack)
                                         (push (list func  gnum) res)
                                         (push #t res)
                                     )
                                     (else
                                        (push  gnum res)
                                     )
                              )
                           )
                           (begin
                             (error 699 gnum)
                             ;;;(push  gnum res)
                           )
                        )
                     )
                   )
                 )
              ((eq? attr 4 )  ;;;;  )
                   (push-#f res)
                  (set! x (pop kstack))
                   ;;;(format #t "pre Right ) ~s~%" x)
                  ;;;New
                  (push (list func x) res)
                  (push #t res)
                   ;;;(format #t " Right res ~s~%" res)
                  )
              ((eq? attr 5 )  ;;;   escape ,  [...]
                  ;;; (format #t "escape func ~s  pat ~s~%" func pat)
                  (set! k (funcall func  pat) )
                    (if  k
                       (push k res)
                    )
                  )
              ((eq? attr 6)   ;;; ignore char in free-space
                 )
              ((eq? attr 7)   ;;;;  comment in free-space
                 (->endline pat)
                 )
              (else (push x res))
        )
        (set! x (R-getc pat))
    )
     (push-#f res)
     res 
))

(define (->endline pat)
  (let* ((x 0))
    (set! x (R-getc pat))
    (while (and (not (eq? x #\newline)) (not (eof-object? x)))
         (set! x (R-getc pat))
    )
))

(define (group-pre pat)
   (let* (( gnum #f)(x 0)(test 0))
           (if (eq? (R-peekc pat) #\?)  ;;; (?  support!)
               (begin
                  (set! x (R-getc pat))  ;;; ?
                  (set! x (R-getc pat))
                   (cond ((eq? x #\:)   ;;;;  No capture group
                                (dec *nocap-num*)
                                 (set! gnum *nocap-num*)
                               )
                         ((eq? x #\<) 
                                 (set! x (R-peekc pat))
                                (cond ( (eq? x #\=)  ;;; modoriyomi
                                       (R-getc pat)  ;;; =
                                       (dec *if-num*)
                                       (set! gnum  (list :backward *if-num*))
                                        )
                                      ( (eq? x #\!)  ;;; modoriyomi not
                                       (R-getc pat)  ;;; !
                                       (dec *if-num*)
                                       (set! gnum  (list :backward  :not *if-num*))
                                        )
                                  (else     ;;; named group
                                    (inc *group-num*)
                                    (set! gnum *group-num*)
                                    (named-group  gnum pat)
                                  )
                                )
                           )
                         ((eq? x #\P)
                                (set! gnum (named-ref pat))
                           )
                          ((eq? x #\?)
                                (set! gnum (regexp-ref pat))
                           )
                         ((eq? x #\=)   ;sakiyomi
                                (dec *if-num*)
                                (set! gnum  (list :forward *if-num*))
                            )   
                          ((eq? x #\!)   ;sakiyomi  not
                                (dec *if-num*)
                                (set! gnum  (list :forward :not *if-num*))
                            )
                          ((eq? x #\>)   ;atomic
                                (dec *if-num*)
                                (set! gnum  (list :atomic *if-num*))
                            )
                          ((or (eq? x #\() (eq? x #\f))  ;if
                              (if (eq? x #\()
                                  (begin
                                    (set! x (R-peekc pat))
                                    (cond ( (char-numeric? x)
                                             (set! test (r-number pat))
                                            )
                                          ((eq? x #\?)  ;;; sakiyomi.modoroyomi
                                             (R-ungetc #\( pat)
                                             (set! test :test)
                                           )
                                          (else
                                              (set! test (name-read pat))
                                             )
                                    )
                                    (when (not (eq? test :test))
                                       (set! x (R-getc pat))
                                        (if (not (eq? x #\))) (error 481 x))
                                    )
                                  )
                                 (begin ;; f( ...)
                                    (set! test (proc-read pat))
                                    ;;;(set! x (R-getc pat))
                                    ;;;(if (not (eq? x #\})) (error 461 x))
                                 )
                              )
                               ;;;;(set! *if-test* test  )
                               (dec *if-num*)
                               (set! gnum  *if-num*)
                               ;;;;  New!!!
                                (set! gnum (list :if *if-num* test))
                            )         
                          (else (set! gnum (g-mode x pat)))
                       )
                    )
                (begin
                   (inc *group-num*)
                   (set! gnum *group-num*)
                )
            )
      gnum
))

(define (proc-read pat)
  (let* ((res ()) (x 0))
    (set! res (read pat))
    res
))
       
(define (named-ref pat)   
    (let* ((name 0)(x 0)(ans 0))
       (set! x  (R-getc pat))
       (cond ((eq? x #\=)
               (set! name (name-read pat))
               (set! ans (list 'group-ref name))
              )
             ((eq? x #\()
                (unread-char x pat)
                ;;;;  (R-ungetc x pat)     ;;;;  problem is here
                (set! name (proc-read pat))  
                (set! ans (list 'group-ref name))
               )
             (else
              (error 489 x)
            )
      )
      (set! x  (R-getc pat))
       ;;;; (format #t "named-ref last x ~s~%" x)
  (if (not (eq? x #\) ) ) (error 488 pat))
  ans
))

(define (regexp-ref pat)  ;;;;; (??{...})  
    (let* ((name 0)(x 0)(ans 0))
       (set! x  (R-getc pat))
       (cond ((eq? x #\{)
               (set! x  (R-peekc pat))
               (if (eq? x #\()
                    (set! name (proc-read pat))  
                    ;;;(set! name (string->symbol (name-read pat)))
                    (set! name (name-read pat))
               )
               (set! ans (list 'nfa-ref name))
              )
             (else
              (error 489 x)
            )
      )
  (if (not (eq? (R-getc pat) #\} ) ) (error 489 pat))
  (if (not (eq? (R-getc pat) #\) ) ) (error 489 pat))
  ;;;;(format #t "regexp-ref ~s~%" ans)
  ans
))


(define (named-group gnum pat)
    (let* ((name (name-read pat))(x 0))
         (set! x (R-getc pat))
     ;;;;(format #t "named-group name ~s x ~s~%" name x)
       (if (not (eq? x #\> ) ) (error 487 pat))
       (set! x (assoc name *named-group*))
       (if  x 
            (set-cdr! x (cons gnum (cdr x)))
            (push (list name gnum) *named-group*)
      )
))

(define (g-mode x  pat)
    (let* ((op #f))
       (when (eq? x #\-)
           (set! op (eq? (R-getc pat) #\i))
       )
       (if (not (eq? (R-getc pat) #\) ) ) (error 387 pat))
       (set! op
          (cond ( op  :case)  ;;; case
                ((eq? x #\i) :nocase)   ;;; nocase
                ((eq? x #\s) :single)
                ((eq? x #\m) :multiple)
                ((eq? x #\n) :normal)
                ((eq? x #\x)  (set! *free-space* #t) 
                   (set-R-table #\space 6 #f)
                   (set-R-table #\newline 6 #f)
                   (set-R-table #\tab  6 #f)
                   (set-R-table #\# 7 #f)
                     #f)
                ((eq? x #\X)  (set! *free-space* #f)
                     (set-R-table #\space 0 #f)
                     (set-R-table #\newline 0 #f)
                     (set-R-table #\tab  0  #f)
                     (set-R-table #\# 0 #f)
                     #f)
                (else (error 411 x))
          )
      )
   (if op (list 'g-mode op) #f)             
))


(define (name-read pat)
   (let* ((x (R-getc pat)) (str ()))
       ;;;;  this may read scheme variable name
     (while  (or (r-word? x)(eq? x #\-)(eq? x #\*)(eq? x #\+)(eq? x #\.))
        (push x str)
        (set! x (R-getc pat))
     )
     ;;;;(format #t "name read ungetc  x ~s~%" x)
    (R-ungetc x pat)
    (string->symbol (list->string (reverse str)))
))

(define (pop-t-f ans)
   (let* ((res ())(x 0) (cont #t))
      (pop ans)  ;;; #t
      (while (and cont (pair? ans))
          (set! x (pop ans))
          (if x
              (push x res)
              (set! cont #f)
          )
       )
   (cons (cons :begin (reverse res)) ans)
))


(define (R-read-stack res) 
  (let* ((ans ()) (x 0)(tmp ()) (new #f) (g-stack ())(f 0))
     (while (pair? res)
        (set! x (R-pop res))
        (set! res (cdr x))
        (set! x (car x))
        (set! new #f)
        (cond ((keyword? x)
                (cond ((eq? x :or)
                       (if (R-left? (car ans))
                            (set! ans (R-pop-group ans))
                            (begin
                               (set! tmp (pop ans))  
                               (if (eq? (car tmp) :begin)
                                  (push (cdr tmp) ans)   ;;;  change -> exp
                                  (push (list  tmp) ans)   ;;;  change -> exp
                               ;;;(set! ans (pop-t-f ans))
                              )
                           )
                       )
                       (push  x ans)
                       (if (R-right? (car res))
                            (begin
                              (set! x (pop res)) 
                              (push (list 'g-top (cadr x)) g-stack)
                              (push x ans)
                           )
                            (begin  ;;; always (car res)=#f
                               (push (pop res) ans)    ;;;  shadow :or by #f
                            )
                       )
                     )
                   (  (or (eq? x :atomic) (eq? x :forward)(eq? x :backward)
                               ) ;;; as :not,  but only group
                        (set! ans (R-pop-group-inside  ans))
                               (set! tmp (pop ans))
                       (push (list x tmp) ans)
                     )
                   ( (eq? x :not)  
                     ;;; (format #t ":not : res ~s ans ~s~%" res ans)
                        (if (R-left? (car ans))
                            (begin
                               (set! ans (R-pop-group ans))
                               (set! tmp (pop ans))
                            )
                            (begin
                               (set! tmp (pop ans))
                               (if (eq? (car tmp) :begin)
                                 (begin
                                  (set! tmp (cdr tmp) )   ;;;  change -> exp
                                  (if (and (pair? (car tmp)) (eq?  (caar tmp)  'back-pos))
                                     (set! tmp (list (car tmp) (pop ans)))
                                   )
                                 )
                                  (set! tmp (list  tmp) )   ;;;  change -> exp
                               ;;;(set! ans (pop-t-f ans))
                              )
                           )
                       )
                       (push (list x tmp) ans)
                     )
                   ( (eq? x :test)
                      ;;;(format #t ":test res ~s ans ~s~%" res ans)
                       (set! f (pop ans))
                        (if (R-left? (car ans))
                            (begin
                               (set! ans (R-pop-group ans))
                               (set! tmp (pop ans))
                            )
                            (begin
                               (set! tmp (pop ans))
                               (if (eq? (car tmp) :begin)
                                 (begin
                                  (set! tmp (cdr tmp) )   ;;;  change -> exp
                                  (if (and (pair? (car tmp)) (eq?  (caar tmp)  'back-pos))
                                     (set! tmp (list (car tmp) (pop ans)))
                                   )
                                 )
                                  (set! tmp (list  tmp) )   ;;;  change -> exp
                               ;;;(set! ans (pop-t-f ans))
                              )
                           )
                       )
                       (push f ans)
                       (push (list x tmp) ans)
                     )
                   ((eq? x :if)
                      ;;;(format #t ":if res ~s ans ~s~%" res ans)
                               (set! tmp (pop ans))
                               (set! tmp (cadr tmp) )   ;;;  if-test
                               (set! tmp (cadr tmp) )   ;;;  only pat of test
                               (when (eq? tmp :test) ;;; sakiyomi
                                   (set! tmp (pop ans))
                               )
                               (set! ans (R-pop-group-inside  ans))
                               (set! x (pop ans))
                               (if (and (null? (cdr x)) (pair? (car x)) 
                                      (eq? (caar x) :or))
                                    (set! tmp (list :if  tmp (cadar x) (caddr (car x)))) 
                                    (set! tmp (list :if  tmp x))
                               )
                       (push  tmp ans)
                     )
                   (else
                     (if  (R-char? (car res))
                             (set! new (list x (list (list 'string (pop res)))))
                            (if (R-right? (car res))
                              (begin
                                 (push x ans)
                            ;;;     R-right!
                                 (set! x (pop res)) 
                                  (push (list 'g-top (cadr x)) g-stack)
                                  (push x ans)
                               )
                               (set! new (list x (list (pop res))))
                           )
                     )
                   )
                 )
                )
              ((and (pair? g-stack)  (equal? (car g-stack ) x))
                  (set! ans (R-pop-group (cons x ans)))
                  (set! x (pop ans))  ;;;  group
                  (set! tmp (pop ans))  ;;;  operation keyword
                  (cond ( (eq? tmp :or)
                            ;;;(format #t ":or  res ~s ans ~s~%" res ans)
                            (set! new (list :or x (pop ans)))
                          )
                         ( (r-range? tmp)    ;;;; :range  :range? :range+
                             (set! new (list tmp  (pop ans) x))
                          )
                         (else
                             (set! new  (list tmp x))
                         )
                  )
                  (pop g-stack)
               )
             ((and (pair? x) (r-range? (car x)))
                        ;;;{n,m}
                  ;;;(format #t "range  x  ~s ans  ~s~% res ~s~%" x  ans res)
                 (set! tmp (pop x))
                 
                 (if  (R-char? (car res))
                             (set! new (list tmp (car x)
                                   (list (list 'string (pop res)))))
                          (begin
                            (push (car x) ans)
                            (push tmp ans)
                            ;;;   always  R-right!
                              ;;;if (R-right? (car res))
                             (set! x (pop res))
                             (push (list 'g-top (cadr x)) g-stack)
                             (push x ans)
                          )
                     )
                )
            ((eq? x #t)
                  (push x ans)
                  (set! ans (pop-t-f ans))
                  (set! new (pop ans))
               )
             ((eq? x #f)
                  (push x ans)
               )
             (else
                (set! new x )
             )
         )
         (when new
           ;;;;(format #t "first new  ~s ans ~s~%" new ans)
            (while (and (pair? ans) (keyword? (car ans)))
                   (set! tmp (pop ans))  ;;;  op
                ;;;(format #t "first new  ~s tmp ~s~%" new tmp)
                   (cond ( (eq? tmp :or)
                             (set! new (list tmp (list new ) (pop ans)))
                           )
                         ( (r-range? tmp) ;;; :range? etc.
                             (set! new (list tmp (pop ans) (list new ) ))
                           )
                         (else
                            (set! new (list tmp (list new)))
                         )
                   )
            )
           ;;;;(format #t "new  ~s ans ~s~%" new ans)
           (push new ans)
        )
         ;;;;(format #t "while  res  ~s ans ~s~%" res ans)
     )
   ans
))


(set-R-table #\. 8 'r-dot)
(set-R-table #\* 1 :*)

(set-R-table #\+ 1  :+)
(set-R-table #\? 1 :?)
(set-R-table #\( 3  'g-top)
(set-R-table #\) 4  'g-end)
(set-R-table #\\ 5 'r-esc)
(set-R-table #\[ 5 'reg-bag)
(set-R-table #\{ 9 'r-range)
(set-R-table #\^ 1  '(r-pos top))   
(set-R-table #\$ 1 '(r-pos end))
(set-R-table #\| 2  :or)

;;;;  make-RL
   (set! R-table RL-table)
   (set-R-table #\\ 5 'r-esc-literal)
   (set! R-table RS-table)

(define (reg-bag pat)
   ;;;(let* ((x 0)(y 0)(z 0)(res '(r-bag)) (cont #t))
   (let* ((x 0)(y 0)(z 0)(res '(r-bag))  (chars ()) (bag ()) (cont #t))
         ;;;(format #t "reg-bag Enter  ~s  ~%" pat)
     (set! x (R-getc pat))
     (when (eq? x #\^)
          (push 'not res)
          (set! x (R-getc pat)) )
     (when (eq? x #\])
          (push x chars)
          (set! x (R-getc pat))
      )
      (if (eq? x #\])
         (begin 
            (push chars res)
            (push bag res)
            (reverse res) 
         )
        (begin
          (if (eq? x #\\)
              (R-ungetc  x pat)
              (push x chars)
          )
          (while (and cont (not (eof-object? x)))
             (set! x (R-getc pat))
              (cond ((eq? x #\-)
                       (if (and (char? (car chars)) (not (eq? (R-peekc pat) #\])))
                              (begin
                                (set! y (char->integer (pop chars)))
                                (set! z (char->integer (R-getc pat)))
                                (push (list '<= y z) bag)
                               )
                              (push x chars)
                      )  )
                   ((eq? x #\])
                      (set! cont #f) )
                   ((eq? x #\\)
                      (set! x (r-esc-bag  pat))
                       (if (char? x)
                          (push x chars)
                           (push x bag)
                       )
                    )
                   (else  (push x chars))
               )
           )
           (push chars res)
            (push bag res)
           (r-bag-red (reverse res ))
       )
     )
))

(define (r-bag-red bag)   ;;;;  [x] -> #\x
   (let*  ((s bag)(top (pop s)) (x (car s)))
       (if (eq? x 'not)
              bag
           (if (and (null? (cadr s)) (= (length x) 1))
               (if (char? (car x))
                      (car x)
                    bag)
               bag
           )
       )
))

                                                                                            
(set! *auto-regexp* make-regexp)



