;;;; 2004.3.25
;;;;   pattern transform
;;;; 4.3
;;;;    template transform
;;;;       if symbol S is not macro,literal,pat, then S is transformed
;;;;;      (template S alias env-assoc (global-value . macro-def))
;;;;        
#|
  trans defitiniton:
  1. symbol 
       macro   => (macro key)
       literal -> (literal word ass)
                   ass is (assoc word env)
       pat variable -> (pat var)
  2. constan
      (), keyword,string,etc.   No change
  3. (P ...   .  Q)
       ->   ((...  (trans P)) .  (trans Q)) 
  4. (P . Q)
       -> ((trans P)  . (trans Q))
  5. #(P1 ...... Pn)
     ->(vector #f (trans P1) .....)
  6. #(P1 ...... Pn ...)
     ->(vector #t (trans P1) .....  (... (trans Pn)))

   (trans P)  will be a pair or not pair(constants).
|#
(define *macro-key* ())
(define *literal* ())
(define *pat-variable* ())
(define *symbol-trans-func* #f)
(define *alias-data* ())
(define *local-env* ())
(define *name-change* ())

 (define (pat-trans p env)
     (set! *symbol-trans-func* pat-symbol-trans)
     (set! *pat-variable* ())
     (pat-trans-common p env)
 )

 (define (template-trans p env)
     (set! *symbol-trans-func* template-symbol-trans)
     (set! *alias-data* ())
     (pat-trans-common p env)
 )


 (define (pat-trans-common  p env)
   (if (pair? p) 
       (pat-pair-trans p env)
       (if (vector? p)
            (pat-vector-trans p env)
           (if (symbol? p) (*symbol-trans-func*  p env)
                p   ;;; constants
           )
       )
   )
)

(define (pat-symbol-trans p env)
   (cond ((eq? p *macro-key*) (list 'macro p))
         ((memq p *literal*) (list 'literal p (assq p env)))
         (else 
           (set! *pat-variable* (cons p *pat-variable*))
           (list 'pat p ))
   )
)

(define (template-symbol-trans p env)
   (cond ((eq? p *macro-key*) (list 'macro p))
         ((memq p *literal*) (list 'literal p ))
         ((memq p *pat-variable*) (list 'pat p ))
         (else
           (let* ((local (assq p env))(res (get-macro p )) (def 0)(flag 0))
             (cond (local
                       ;;;(list 'local  p (gen-alias p) local))
                       ;;; alias must have same value
                      (set! res (gen-alias p))
                       (symbol-value-set res (cdr local)) 
                       (list 'local  p res local))
                   ((special? p)
                          ;;;  for special-syntax this must before syntax macro
                          ;;; Don't use eval, because eval call macro-expand-rec!!!
                           ;;; (list 'special p (symbol-value p ))
                           (list 'special p (special-alias p )))  ;; 04.5.7
                   (res   ;;; macro
                         (set! def (car res))
                         (set! flag (caddr def))
                         (if (or (eq? flag #f) (eqv? flag 0))
                               ;;; p is alias
                             (list 'syntax-macro  (cadr def) p res) 
                             (list 'syntax-macro p (cadr def) res) 
                         )
                      )
                   (else 
                      ;;;(list 'template p (gen-alias p) )
                      ;;;; template may be a function
                        (set! res (gen-alias p))
                        (symbol-value-set res (cons :MACRO_ALIAS  p)) ;;;
                           ;;; 04.5.3  this needs for compiler, p may be defined in
                           ;;; another file
                      (if (symbol-bind? p)
                          ;;;(symbol-value-set res (symbol-value p)) 
                          ;;;;  p may be already alias! let-syntax used
                          ;;;;    in a define-syntax template!
                          ;;;;  04.4.23
                        (begin
                          (set! def #t)
                          (while (and def (symbol-bind? p))
                            (set! def (symbol-value p))
                            (if (and (pair? def) (eq? (car def) :MACRO_ALIAS))
                                (begin
                                  (set! p (cdr def))
                                )
                               (set! def #f)
                            )
                          )
                          (symbol-value-set res (cons :MACRO_ALIAS  p)) 
                        )
                                  ;;;;  04.4.18 for compiler
                      )
                      ;;(format #t "template p ~s res ~s value ~s~%" p res 
                      ;;     (symbol-value res))
                      (list 'template p res)
                   )
             )
           )   )
   )
)
;;;;    alias  should be same in a template!!! 
(define (gen-alias s)
  (let* ((a (assq s *alias-data*)))
     (if a (cdr a)
        (begin
          (set! a (gensym))
          (set! *alias-data* (cons (cons s a) *alias-data*))
          a
        )
     )
))


(define (pat-pair-trans p env)
   (let* ((top (car p)) (rest (cdr p)))
     (if (pair? rest)
        (if (eq? (car rest) '...)
             (cons (list '... (pat-trans-common  top env)) (pat-trans-common (cdr rest ) env))
             (cons (pat-trans-common  top env) (pat-trans-common rest env))
        )
       (cons (pat-trans-common  top env) (pat-trans-common rest env))
       ;;; (list (pat-trans-common  top env))      wrong for (f . x)!!!
    )
))

(define (pat-vector-trans p env)
   (let* ((n (vector-length p)) (res ()) (flag #f)(i 0)(q 0))
      (set! n (- n 1))
      (while (> n 0)
         (push (pat-trans-common  (vector-ref p i) env) res)
         (set! n (- n 1))
         (set! i (+ i 1))
      )
      (if (= n 0)
        (begin
            (set! q (vector-ref p i))
            (if (eq? q '...)
               (begin
                 (set! q (pop res))
                 (push (list '... q) res)
                 (set! flag #t)
                )
                 (push (pat-trans-common  q env) res)
              ) 
          )
      )
      (set! p (cons flag (reverse res)))
      (cons 'vector p)
))
              
   
#|
 (define *macro-key* 'cond)
 (define *literal* (list 'else))

 (pat-trans '(cond (else y ...)) ())

 (pat-trans '(cond (test y ...) clause ...) ())

 (define *macro-key* 'do)
 (pat-trans '(do ((var init step ...) ...)
         (test expr ...)
         command ...)   ())   

 (pat-trans '(do ( #(var init step ...) ...)
         (test expr ...)
         #(foo bug)
         ()
         #()
         command ...)   ())
  (template-trans ' (letrec
       ((loop
           (lambda (var ...)
             (if test
                 (begin
                   (if #f #f)
                   expr ...)
                 (begin command ...
                   (loop (do "step" var step ...)
                       ...)
                 )
              )
            )
           ))
        (loop init ...))      ())
 
|#

