;;; make-pdf.ss
;;;  2012.12.14
;;;   2013.12.23 verbatim error fix
;;;
(define Head-sub
"\\documentclass[10pt]{jarticle}
\\topmargin = 0cm
\\oddsidemargin = 0cm
\\textwidth = 16cm
\\textheight = 23cm

\\usepackage{enumerate}
\\usepackage{amsthm}
\\usepackage{multicol}
\\newtheorem{thm}{Theorem}
\\newtheorem{lem}{Lemma}

\\newcommand{\\disp}{\\displaystyle}

\\pagestyle{empty}

\\begin{document}
")
(define (make-Head)
  (let * ((font "" ))
      (if Tex-multi (set! font "\\fontsize{8pt}{14pt}\\selectfont"))
      (tk-string 
        Head-sub font #\newline )
))

(define Tail
"
\\end{document}
")

(define (get-ext file)
  (let* ((y (file-sepa-ext file )))
       (cdr y)
))


(define (make-temp-tex file)
  (let* ((x 0)(y (file-sepa-ext file )) (base (tk-string "temp-" (car y)))  (ext (cdr y)))
     (while (file-exists? (tk-string base x ".tex"))
       (inc x)
     )

   (cons      (tk-string base x ".tex")
     (tk-string base  x)  )
))

(define (out-str str out)
    (format out "~a" str)
)

(define (tab->space x)
  (let* ((ind 0) (tab (string #\tab)))
     (set! ind (string-index tab x))
    (if ind 
       (string-append (substring x 0 ind) "    "
            (tab->space (substring x (+ ind 1) :end)))
        x
    )
)) 
       
(define (verbatim x)
   (let* ((h  33)     ;;; #\!
           (s 0))
       (while (string-index (string (integer->char h)) x)
           (inc h)
           (when (> h 126)  ;;;;error
               (format #t "Too complex text!!!  exit!~%~a~%" x)
               (bye)
           )
       )
      (set! h (integer->char h))
      (tk-string 
           "\\end{verbatim}" #\newline
           "\\verb" h x h #\newline
           "\\begin{verbatim}" #\newline  )
))

(define (check-verbatim x)
   (let* ((verb "\\end{verbatim}")
          (ind (string-index verb x)) )
      (if ind 
       (cons #t (verbatim x))
        (cons #f x)
     )
))

(define (count-space x)
;;;  return count of space in head of x
 (let* ((c 0))
   (while (eq? (string-ref x c) #\space)
      (inc c)
   )
 c
))

 
(define (out-file file out)
  (let* ((in (open-input-file file)) (x (read-line in))(MAXL 85) (fx 0)(top 0)
          (indent ""))
    ;;; (if Tex-multi (set! MAXL 55))  ;;; for 10pt scriptsize
    (if Tex-multi (set! MAXL 45)) ;;;  for 8pt
    (while (not (eof-object? x))
       (set! x (tab->space x))  ;;; tab-> "    "
       (set! fx (check-verbatim x))
       (set! x (cdr fx))
    
         ;;;;  if verb+ ...+  , this must be single line!!!
         (when (not (car fx))  ;;; not verb
            (when (> (string-length x) MAXL)
                (set! top (count-space x))
                (if (> top 16) (set!  top 16))
                (if (< top 2) (set!  top 8))
               (set! indent (make-string top #\space))
            )
            (while (> (string-length x) MAXL)
                (format out "~a~%" (substring x 0 MAXL))
                (set! x (string-append indent (substring x MAXL :end)))
               ;;;    (set! x (substring x MAXL :end))
            )
         )
       (format out "~a~%" x)
       (set! x (read-line in))
    )
   (close-port in)
))

(define (action file)
  (let* ((y (file-sepa-ext file ))
            (base (car y)) (ext (cdr y)))
     (cond  ((equal? ".tex" ext) (tex file base))
            ((equal? ".pdf" ext) )
            ((equal? ".dvi" ext) (dvi->pdf file))
            (else  (ppp file))
     )
))

(define (tex file base)
    (system (tk-string "platex " file))
     (system (tk-string "platex " file))
     (system (tk-string "dvipdfmx  "  base ".dvi"))
)
(define (dvi->pdf file)
     (system (tk-string "dvipdfmx  "  file))
)

(define Tex-multi #f)

(define (ppp file)
   (let* ((x 0) (temp2 (make-temp-tex file))
              (temp (car temp2))(base (cdr temp2)) (dvi (tk-string base ".dvi")) (out (open-output-file temp)))
    ;;; (system (tk-string "cat Templ-head " file " Templ-tail>" temp)) 
    (out-str (make-Head) out)
      (when Tex-multi
         (format out "\\begin{multicols}{2}~%")
         ;;;(format out "\\scriptsize~%")
      )
     (format out "\\begin{verbatim}~%")
    (out-file file out)
     (format out "\\end{verbatim}~%")
      (when Tex-multi
         (format out "\\end{multicols}~%")
      )
    (out-str Tail out)
    (close-port out)
    (tex temp base)
   ;;(remove-file temp)
))

(let* ((f #f))
  (when (not (pair? *argv*))
     (format #t "usage:  makepdf [option] file1 ...~%")
     (format #t "        *.tex   -> platex,platex->  pdf~%")
     (format #t "        *.dvi   ->   pdf~%")
     (format #t "    other[default] ->  temp-xxxxx.tex ,  temp-xxxxxx.pdf~%")
     (format #t "  option:~%")
     (format #t "         -m   two-column  output temp....pdf~%")
     (format #t "         -s   [default]  one-column output temp....pdf~%")
     (bye)
  )
 (while (pair? *argv*)
   (set! f   (pop *argv*))
   (cond ( (equal? f "-m")
           (set! Tex-multi #t))
         ( (equal? f "-s")
           (set! Tex-multi #f))
         (else (action  f))
  )
 )
)

(bye)
