;;;  or    as a syntax transformer
;;;  2010.8.10
#|
(define define-syntax
   (->syntax  (lambda (x)
      (let* ((dummy (car x)) (para (cdr x)) (key (car para)) (exp (cdr para)))
       `(define ,key
          (->syntax ,@exp) )
      )
     )
   )
)

(define define-syntax
   (macro (key exp)
         `(define ,key
             (->syntax (begin (set! *macro-key* ',key) ,exp) )
          ) 
   )
)
|#

(define-syntax or 
    (lambda (x)
      (let* ((key (car x)) (para (cdr x))) 
         (if (null? para) #f
            (if (null? (cdr para))
                 (car para)
                 (let* ((a (gensym)))
                    `(let* ((,a ,(car para)))
                       (if ,a ,a
                           (,key ,@(cdr para))
                       )
                     )
                  )
             )
         )
      )
    )
)
;;;
(define-syntax and 
   (lambda (x)
      (let* ((key (car x)) (para (cdr x)))
         (if (null? para) #t
            (if (null? (cdr para))
                 (car para)
                 (let* ((a (gensym)))
                    `(let* ((,a ,(car para)))
                       (if ,a 
                           (,key ,@(cdr para))
                           ,a
                       )
                     )
                  )
             )
         )
      )
    )
)

;;;
(define-syntax when
   (lambda (x)
      (let* ((para (cdr x)) (cond (car para)) (body (cdr para)))
        `(if ,cond (begin ,@body))
      )
     )
)
;;;
(define-syntax push
   (lambda (x)
      (let* ((para (cdr x)) (obj (car para)) (var (cadr para)))
        `(begin
             (set! ,var (cons ,obj  ,var))
             ,var
         )
      )
     )
)
;;;
(define-syntax pop
   (lambda (x)
      (let* ((a (gensym)) (var (cadr x)))
        `(let* ((,a (car ,var)))
             (set! ,var (cdr  ,var))
             ,a
         )
      )
     )
)
;;;
(define-syntax while
   (lambda (x)
      (let* ((para (cdr x)) (cond (car para)) (body (cdr para))
              (loop-f (gensym)))
         `(letrec  ((,loop-f (lambda ()
                        (if ,cond
                           (begin ,@body (,loop-f)) )
                        )))
               (,loop-f)
          )
       )
     )
)
;;;;
#|
(define-macro (cond cond-x . cond-body)
     (let* ()
        (if (null?  cond-body)
          (begin  ;;;  this is last
              (if (or (eq? (car cond-x) 'else) (eq? (car cond-x ) #t))
                  (if (pair? (cdr cond-x))
                             (cons 'begin (cdr cond-x))
                             (make-unspecified)
                  )
                ;;;`(if ,(car cond-x) ,(cadr cond-x))
                `(if ,(car cond-x)
          ,(if (pair? (cdr cond-x)) (cadr cond-x) (make-unspecified)) )
               )
          )
         (begin
         `(if ,(car cond-x) ,(if (pair? (cdr cond-x))
               (cons 'begin (cdr cond-x)) (make-unspecified))
                (cond  ,@cond-body)
            )
          )
       )
))
|#

(define-syntax cond
   (lambda (x)
     (let* ((para (cdr x)) (cond-x (car para)) (cond-body (cdr para) ))
        (if (null?  cond-body)
          (begin  ;;;  this is last
              (if (or (eq? (car cond-x) 'else) (eq? (car cond-x ) #t))
                  (if (pair? (cdr cond-x))
                             (cons 'begin (cdr cond-x))
                             (make-unspecified)
                  )
                ;;;`(if ,(car cond-x) ,(cadr cond-x))
                `(if ,(car cond-x)
          ,(if (pair? (cdr cond-x)) (cadr cond-x) (make-unspecified)) )
               )
          )
         (begin
         `(if ,(car cond-x) ,(if (pair? (cdr cond-x))
               (cons 'begin (cdr cond-x)) (make-unspecified))
                (cond  ,@cond-body)
            )
          )
       )
)))

