;;;; read table utility
;;;;  2004.5.20
;;;;
(in-package "read_table")
(define TSIZE 128)
(define sys-table  *read-table*)

(define (get-syntax c  table)
   (let* ((n (char->integer c))(data 0))
      (if (< n TSIZE)
           (begin
               (vector-ref table n)
           )
          (error E_range c)
      )
 ))


(define constituent (car (get-syntax #\a sys-table)))
(define w_space     (car (get-syntax #\space sys-table)))
(define s_escape     (car (get-syntax #\\ sys-table)))
(define m_escape     (car (get-syntax #\| sys-table)))
(define t_macro     (car (get-syntax #\(  sys-table)))
(define non_t_dispatch_macro     (car (get-syntax #\#  sys-table)))
(define dispatch_macro     (/ #x240 4))
(define illegal_char  (/ #x8000 4))
(define non_t_macro     (/ #x140 4)) 

(define *type-of-char*
   `( (,constituent . constituent) (,w_space . w_space)
      (,s_escape . s_escape) (,m_escape . m_escape)
      (,t_macro . t_macro) (,non_t_macro . non_t_macro)
      (,dispatch_macro . dispatch_macro) (,illegal_char . illegal)
      (,non_t_dispatch_macro . non_t_dispatch_macro) ))


(define (is_macro type)
    (= (logand type t_macro) t_macro))

(define (is_dispatch type)
    (= (logand type dispatch_macro) dispatch_macro))


(define (set-syntax c type func table)
   (let* ((n (char->integer c))(data 0))
      (if (< n TSIZE)
           (begin
              (set! data  (vector-ref table n))
              (set-car! data type)
              (set-cdr! data func)
           )
          (error E_range c) 
      )
 ))


(define (make-dispatch c type table)
   (let* ((n (char->integer c)) (data 0))
      (if (< n TSIZE)
           (begin
              (set! data  (vector-ref table n))
              (set-car! data type)
              (set-cdr! data (make-vector TSIZE))
           )
          (error E_range c)
      )
 ))

(define (set-dispatch c subc func table)
   (let* ((n (char->integer c)) (data 0))
      (if (< n TSIZE)
          (begin
              (set! data  (vector-ref table n))
              (if  (not (is_dispatch (car data)))
                    (error E_illegal))
              (set!  data (cdr data))
              (set! n (char->integer subc))
              (if (>= n TSIZE)
                     (error E_range c))
              (vector-set! data n  func)
             
           )
          (error E_range c)
      )
 ))

(define (get-dispatch-table  c  table)    ;;;; return dispatch table
   (let* ((n (char->integer c)) (data 0))
      (if (< n TSIZE)
          (begin
              (set! data  (vector-ref table n))
              (if  (not (is_dispatch (car data)))
                    (error E_illegal))
              (cdr data)
           )
          (error E_range c)
      )
 ))

(define (get-dispatch-macro c subc table)  ;;;  return macro function
   (let* ((n (char->integer c)) (data 0))
      (if (< n TSIZE)
          (begin
              (set! data  (get-dispatch-table c table ))
              (set! n (char->integer subc))
              (if (>= n TSIZE)
                     (error E_range c))
              (vector-ref data n )
             
           )
          (error E_range c)
      )
 ))

(define (make-read-table )
  (let* ((table (make-vector TSIZE)) (del (- TSIZE 1)) 
              (type 0)(space (char->integer #\space)))
    (do ((i 0))
        ((= i TSIZE) table)
        (if (= i del)
             (set! type w_space)
             (if (<= i  space)
                  (set! type w_space)
                  (set! type constituent)
             )
         )
        (vector-set! table i (cons type (make-undef)))
        (set! i (+ i 1))
   )
))

(define (copy-read-table table)
   (let* ((newtable (make-vector TSIZE)) 
              (type 0)(data 0) (func 0)(undef (make-undef)))
    (do ((i 0))
        ((= i TSIZE) newtable)
        (set! data (vector-ref table i))
        (set! type (car data))
        (set! func (cdr data))
        (if (vector? func) (set! func (vector-copy func)))
        (vector-set! newtable i (cons type func))
        (set! i (+ i 1))
   )
))

;;;  keep original in sys:
(define sys:*read-table* *read-table*) 
;;;  use copy
(set! *read-table* (copy-read-table *read-table*))

#| example
;;;;   #'name  -> (function name)
(define (rm-func path ch)
   ;;;   ch='    
  (list 'function (read path))
)
(set-dispatch #\# #\'  rm-func *read-table*)


;;;; {...}  -> (begin  ...)
(define *read-flag* #t)

(define (rm-end path)
   (set! *read-flag*  #f)
   #f   ;;; must return not undef
)

(define (rm-begin path)
  (let*  ((res ()) (x 0))
     (push 'begin res)
     (set! *read-flag* #t)
     (while *read-flag*
         (push (read path) res)
     )
    (set! *read-flag* #t)
    (reverse (cdr  res))  ;;;; drop #f
))

(set-syntax #\} t_macro rm-end *read-table*)
(set-syntax #\{ t_macro rm-begin *read-table*)
|#

#|

 (do ((i 0 (+ i 1))(c 0)) ((= i 128)) 
       (set! c (integer->char i))
       (format #t "~s  ~s  ~s    " i c
             (cdr (assq  (car (get-syntax c *read-table*)) *type-of-char*))  )
       (if (= (modulo i 4) 3) (newline))
 )      
|#

 
  
