;;;   socket.ss
;;;  when-socket-ready
;;;  find-netport
;;;  net-run
;;;  2004.7.16
;;;;

(in-package "socket")
(define *waiting-ports* ())
(define *polling* -1)
(define *time-interval* 500)

(define (when-socket-ready port func)
  (let* ((pf 0))
    (if (eq? func #f)
      (begin
        (set! pf (assq port *waiting-ports*))
        (if  pf 
            (set! *waiting-ports* (remove pf *waiting-ports*))
        )
      )    
      (begin
        (push (cons port func) *waiting-ports*)
        (if  (negative?  *polling* )
             (start-polling)
        )
      )
   )
))

(define (start-polling)
   (set! *polling* (after *time-interval* ports-polling))
)

(define (ports-polling)
    (let* ((all *waiting-ports*)(job 0)(port 0))
      (while (pair? all)
        (set! job (pop all))
        (set! port (car job))
        (and  (not (closed-port? port))  (char-ready? port)
            (funcall (cdr job))
        )
      )
   (if  (pair?  *waiting-ports*)
         (set! *polling* (after *time-interval* ports-polling))
         (set! *polling*  -1)
   )
)) 
;;;;  find unused netport
(define (find-netport)
  (let* ((net #f))
     (do ((i 52000 (+ i 1)))
         (net net)
         (set! net (netport i))
         (if (netport-used?  net) 
               (set! net #f)
               (format *stdmsg* "port number ~s~%" i)
         )
         (if (> i 52999) (error E_socket i))
     )
))

;;;; Net communication wth unix command (or  C program)
(define (net-run  . cmds)
  (let* ((listen #f) (client 0)(stream 0)(net 0)(pid -1) (pr 0))
    (while (not listen)
       (set! net (find-netport))
      (catch (set! listen  (socket-listen net)))
      (when listen
        (set!  client  (socket-connect net))
        (set! stream   (socket-accept  listen))
        (close-port listen)
        (set! pr (map tk-string cmds))
        (set! pr (list (cons 'run-process pr) client client #f))
        (set! pid (apply redirect pr))
        (wait-netrun pid client)
       )
    )
   (cons pid  stream)
))

(define (wait-netrun  pid client )
   (if (process-exited? pid)
         (close-port client) ;;; then we will get eof-object!!!
         (after 5000 (lambda ()(wait-netrun pid client)))
  )
)
;;;;
 
(export 'when-socket-ready)
(export 'find-netport)
(export 'net-run)
(use-package "socket" "ss")

