;;;  Game 
;;;
;;;  game is (left-set .  right-set)
(define (left-of x)
  (if (pair? x)
        (car x)
        ()
))

(define (right-of x)
  (if (pair? x)
        (cdr x)
        x
))

(define (gmake left right)
    (cons left right)
)

(define (print-game x)
  (cond ((game= x Zero) (write 0))
        ((game= x One)  (write 1))
        ((game= x Star)  (display  #\* ))
        ((game= x Up)  (display  "U" ))
        ((game= x Down)  (display  "D" ))
        (else
            (write-char #\{)
            (map print-game (left-of x))
            (write-char #\space)
            (write-char #\|)
            (write-char #\space)
            (map print-game (right-of x))
            (write-char #\})
            (write-char #\space)
       )
 )
)

;;;  args splitted by #\|

(define (make-game . args)
  (let* ((left ()) (right ()))
     (do ((x 0))
         ((or (eq? args ()) (eq? x #\|)))
         (set! x (pop args))
         (if (not (eq? x #\|))
             (set! left (add-element x left))
         )
     )
     (do ((x 0))
         ( (eq? args ()) )
         (set! x (pop args))
         (set! right (add-element x right))
     )
    (gmake left right)
))


(define Zero (make-game))
(define One (make-game  Zero #\| ))
(define Star (make-game  Zero #\|  Zero))
(define Up (make-game  Zero #\|  Star))
(define Down (make-game  Star #\|  Zero))


(define (negative game)
  (let* ((left (left-of game)) (right (right-of game)))
     (set! game (map negative right))
     (set! right (map negative left))
     (gmake game right)
))

(define (sum x y)
  (let* ((xl (left-of x)) (xr(right-of x))
         (yl (left-of y)) (yr (right-of y)) (left ()) (right ()))
    (set! left (map (lambda (x) (sum x y) )  xl))
    (set! left (set-union left (map (lambda (y) (sum x y) )  yl)))
    (set! right (map (lambda (x) (sum x y) )  xr))
    (set! right (set-union right (map (lambda (y) (sum x y) )  yr)))
     (gmake left right)
))

(define -One (negative One))
(define Two (sum  One  One))
(define G1/2  (make-game  Zero #\| One ))


;;; P.78
#|    --->  set.c
(define (forall-element  set test)
   (do ((ans #t) (x 0))
      ((null? set) ans)
      (set! x (pop set))
      (when (not (funcall test x))
          (set! set ())
          (set! ans #f)
      )
   )
)
|#  
(define (game>=  g h)
   (and 
      (forall-element  (right-of g)  (lambda(x) (not (game>= h x)))  )
      (forall-element  (left-of h)  (lambda(x) (not (game>=  x g )))  )
   )
)

(define (game<= g h)
    (game>= h g)
)

(define (game= g h)
  (and (game>= g h) (game>= h g))
)

;;;   ||    fuzzy?
(define (game// g h)
  (and (not (game>= g h) )(not (game>= h g)))
)
;;;   ||>    fuzzy or  >
(define (game//>  g h)
  (not (game>=  h g) )
)

;;;   ||<    fuzzy or  <
(define (game//>  g h)
  (not (game>=  g h) )
)

;;;P.81-82
(define (game-is-number?  g )
   (and
      (forall-element (left-of g) game-is-number? )
      (forall-element  (right-of g) game-is-number? )
      (forall-element  (left-of g)
           (lambda (x)  (forall-element  (right-of g)
                              (lambda (y) (not (game>= x y)) )) )
      )
   )
)

;;;P.109   simplifying
;;;   there are some BUG
;;;  Mr. Takano says
;;;  the following remove must be remove-1
;;;;   because  if use remove all entry's may be  removed!!!!
;;;;   So , remove-1 must be used.
;;;;    MADA !!!
;;;;


(define (left-dominated-simplify g)
   (let* ((memo (left-of g))(x 0)(y 0)(left (left-of g)))
     (while (pair? left)
        (set! x (pop left))
        (set! y (exists-element left (lambda (u) (game>= u x))))
        (if y  ;;;;  xis dominated by y
           (set! memo (remove x memo))
           (begin     
              (set! y (exists-element left (lambda (u) (game<= u x))))
              (when y   ;;; y is dominated by x
                 (set! memo (remove y memo))
                 (set! left (remove y left))
              )
          )
        )
     )
   (if (equal? memo (left-of g))
        g
        (gmake memo (right-of g))
   )
))

(define (right-dominated-simplify g)
   (let* ((memo (right-of g))(x 0)(y 0)(right (right-of g)))
     (while (pair? right)
        (set! x (pop right))
        (set! y (exists-element right (lambda (u) (game<= u x))))
        (if y  ;;;;  xis dominated by y
           (set! memo (remove x memo))
           (begin
              (set! y (exists-element right (lambda (u) (game>= u x))))
              (when y   ;;; y is dominated by x
                 (set! memo (remove y memo))
                 (set! left (remove y left))
              )
          )
        )
     )
   (if (equal? memo (right-of g))
        g
        (gmake (left-of g) memo)
   )
))
   ;;;;  reversible 
(define (left-reversible x g)  ;;;x if one of Left of G
    (exists-element (right-of x) (lambda(x) (game<= x g)))
)

(define (left-reversible-simplify g)
  (let* ((memo (left-of g))(left (left-of g)) (x  0)(y 0))
     (while (pair? left)
        (set! x (pop left))
        (set! y (left-reversible x g))
        (when y
           (set! memo (remove x memo))
           (set! memo (set-union  memo (left-of y)))
        )
     )
    (if (equal? memo (left-of g))
        g
        (gmake memo (right-of g))
   )
))

(define (right-reversible x g)  ;;;x if one of right of G
    (exists-element (left-of x) (lambda(x) (game>= x g)))
)

(define (right-reversible-simplify g)
  (let* ((memo (right-of g))(right (right-of g)) (x 0)(y 0))
     (while (pair? right)
        (set! x (pop right))
        (set! y (right-reversible x g))
        (when y
           (set! memo (remove x memo))
           (set! memo (set-union  memo (right-of y)))
        )
     )
    (if (equal? memo (right-of g))
        g
        (gmake (left-of g) memo)
   )
))

(define (simplify g)
  (let* ((memo g)(cont #t))
     (if (null? g) 
         g
       (begin
         (set! g (gmake (map simplify (left-of g))
                       (map simplify (right-of g))))
         (set! memo g) 
         (while cont
            (set! g (left-dominated-simplify g))  
            (set! g (right-dominated-simplify g))  
            (set! g (right-reversible-simplify  g))  
            (set! g (left-reversible-simplify  g))  
            (set! cont (not (eq? g memo)))
            (set! memo g)
         )
         g
       )
     )
))
 
#| example 
    (define y (sum G1/2 G1/2))
    (simplify y)

    (define z (make-game G1/2 #\|))   ;;;  = 1
    (simplify z)

    (define z (make-game Up #\| Up))   ;;;  = {0 1 Up}
    (simplify z)
 

|#


