package require Tkogl

pack [OGLwin .gl -width 300 -height 300 ] -side left -expand yes -fill both


# Set up default gl values
eval .gl eval \
	-matrixmode projection \
	-loadidentity \
	-ortho -5 5 -5 5 -5 5\
	-material frontandback ambientanddiffuse 0.7 0.7 0.7 \
	-material frontandback specular 1 1 1 \
	-material frontandback shininess 100 \
	-enable autonormal \
	-shademodel smooth \
	-enable light0 \
	-enable lighting \
	-enable normalize \
	-enable depthtest

#
# Set up default nurbs values
#

# Init nurbs surface data
set cpoints {}
for {set u 0} {$u < 4} {incr u} {
    for {set v 0} {$v < 4} {incr v} {
	lappend cpoints [expr ($u-1.5)*2.0] [expr ($v-1.5)*2.0] 
	if {($u==1||$u==2)&&($v==1||$v==2)} {
	    lappend cpoints 3
	} else {
	    lappend cpoints -3
	}
    }
}
set knottype "uniform"
set order 4
set culling f
set sampling 50
set displaymode fill

# Create control point mesh display
proc cpoint { i j } {
    # returns control point (i,j) 
    global cpoints
    set k [expr ($i*4+$j)*3]
    return [lrange $cpoints $k [expr $k+2]]
}
for {set i 0} {$i < 4} {incr i} {
    lappend l -begin linestrip 
    for {set j 0} {$j < 4} {incr j} {
	append l " -vertex [cpoint $i $j]"
    }
    lappend l -end
}
for {set i 0} {$i < 4} {incr i} {
    lappend l -begin linestrip 
    for {set j 0} {$j < 4} {incr j} {
	append l " -vertex [cpoint $j $i]"
    }
    lappend l -end
}
set cpointmesh [eval .gl newlist $l]
set cpointdisplay [.gl newlist -call $cpointmesh]
set cpointflag 1

# Compute the nurbs picture
proc computenurbs {} {
    global cpoints sampling culling displaymode nurbs order knottype cpointdisplay
    catch {.gl deletelist $nurbs}
    set knots {}
    set n [expr $order+4]
    if {$knottype=="uniform"} {
	set k [expr 1.0/($n-1)]
	for {set i 0} {$i<$n} {incr i} {
	    lappend knots [expr $k*$i]
	}
    } else {
	set j [expr $n-$order*2]
	set d [expr 1.0/($j+1)]
	for {set i 0} {$i<$j} {incr i} {
	    lappend knots [expr ($i+1)*$d]
	}
	for {set i 0} {$i<$order} {incr i} {
	    lappend knots 1
	    set knots [linsert $knots 0 0]
	}
    }
    set nurbs [eval .gl nurbssurface \
	    -sampling $sampling -culling $culling -uorder $order \
	    -vorder $order -displaymode $displaymode\
	    -vknots $knots -uknots $knots -controlpoints $cpoints]
    .gl main -clear colorbuffer depthbuffer \
	    -enable lighting -linewidth 1 -call $nurbs\
	    -disable lighting -color 1 1 0 -linewidth 2 -call $cpointdisplay
}
computenurbs
   
# Set up rotation bindings
set xrot 0
set yrot 0
set pmode fill

bind .gl <Button-1> { set x %x ; set y %y }
bind .gl <B1-Motion> { 
    incr xrot [expr %x - $x] 
    incr yrot [expr %y - $y] 
    set x %x
    set y %y
    .gl eval -matrixmode modelview \
            -loadidentity \
            -rotate $yrot 1 0 0\
            -rotate $xrot 0 1 0
    .gl redraw
}

#
# Set up rest of interface
#
pack [frame .cmd] -side right -fill y

# Set up Command buttons
pack [frame .cmd.button] -fill x
pack [button .cmd.button.quit -text "Quit" -command exit] -fill x -side left
pack [button .cmd.button.recompute -text "Recompute" -command computenurbs]\
	-fill x -side left
pack [checkbutton .cmd.button.showframe -text Frame -var cpointflag -command {
    if $cpointflag {
        .gl newlist $cpointdisplay -call $cpointmesh
    } else {
        .gl newlist $cpointdisplay
    }
    .gl redraw
}] -fill x -side left
        
# Set up display mode interface
pack [frame .cmd.dmode] -fill both
pack [label .cmd.dmode.label -text "Display Mode"] -side left
pack [frame .cmd.dmode.value -bd 2 -relief sunken] -side right
foreach dmode {fill outlinepolygon outlinepatch} {
    pack [radiobutton .cmd.dmode.value.$dmode -text $dmode -var displaymode \
        -value $dmode -anchor w -command "computenurbs;.gl redraw"] \
        -fill x
}

# Set up sampling tolerance interface
pack [frame .cmd.stol] -fill both
pack [label .cmd.stol.label -text "Sampling Tolerance"] -side left
pack [frame .cmd.stol.value -bd 2 -relief sunken] -side right
foreach stol {10 25 50} {
    pack [radiobutton .cmd.stol.value.$stol -text $stol -var sampling\
	    -value $stol -anchor w -command "computenurbs;.gl redraw"] \
	    -fill x
}

# Set up knot type interface
pack [frame .cmd.ktype] -fill both
pack [label .cmd.ktype.label -text "Knot Distribution"] -side left
pack [frame .cmd.ktype.value -bd 2 -relief sunken] -side right
foreach ktyp {uniform non-uniform} {
    pack [radiobutton .cmd.ktype.value.$ktyp -text $ktyp -var knottype\
	    -value $ktyp -anchor w -command "computenurbs;.gl redraw"] \
	    -fill x
}

# Set up Order interface
pack [frame .cmd.order] -fill both
pack [label .cmd.order.label -text "Polynomial Order"] -side left
pack [frame .cmd.order.value -bd 2 -relief sunken] -side right
foreach ord {2 3 4} {
    pack [radiobutton .cmd.order.value.$ord -text $ord -var order\
	    -value $ord -anchor w -command "computenurbs;.gl redraw"] \
	    -fill x
}


